#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from oscoroutines._base_task import base_task
import inspect

class inline_task(base_task):

	"""
	desc:
		A task controls the coroutine for a Python generator function.
	"""

	def __init__(self, coroutines, function_name, python_workspace, start_time,
		end_time):

		"""
		desc:
			Constructor.

		arguments:
			function_name:
				desc:	The name of a Python generator function.
				type:	str
			python_workspace:
				desc:	The python-workspace object.
				type:	python_workspace
		"""

		base_task.__init__(self, coroutines, start_time, end_time)
		self.function_name = function_name
		self.python_workspace = python_workspace

	def launch(self):

		"""See base_task."""

		if self.function_name not in self.python_workspace._globals:
			raise osexception(u'"%s" is not defined' % self.function_name)
		self.coroutine = self.python_workspace._globals[self.function_name]
		if not inspect.isgeneratorfunction(self.coroutine):
			raise osexception(
				u'"%s" is not a generator function' % self.function_name)
		self.coroutines.event('launching %s' % self.coroutine)
		self.coroutine = self.coroutine()
		self.coroutine.send(None)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from openexp._keyboard.keyboard import Keyboard
from psychopy import event
import pyglet.window.key
from openexp.backend import configurable


class Psycho(Keyboard):

	"""
	desc:
		This is a keyboard backend built on top of PsychoPy.
		For function specifications and docstrings, see
		`openexp._keyboard.keyboard`.
	"""

	# The keymap is an incomplete attempt at translating keys from the PyGame
	# names to the names used by PsychoPy
	keymap = {
		'!' : 'exclamation',
		'"' : 'doublequote',
		'#' : 'hash',
		'$' : 'dollar',
		'&' : 'ampersand',
		'\'' : 'quoteleft',
		'(' : None,
		')' : None,
		'*' : 'asterisk',
		'+' : 'plus',
		',' : 'comma',
		'-' : 'minus',
		'.' : None,
		'/' : 'slash',
		':' : 'colin',
		';' : 'semicolon',
		'=' : 'equal',
		'>' : 'greater',
		'?' : 'question',
		'@' : 'at',
		'[' : 'bracketleft',
		'\\' : 'backslash',
		']' : 'bracketright',
		'^' : None,
		'_' : 'underscore'
		}

	def __init__(self, experiment, **resp_args):

		Keyboard.__init__(self, experiment, **resp_args)
		self._latest_modifiers = {}

	def valid_keys(self):

		l = []
		for i in dir(pyglet.window.key):
			if isinstance(getattr(pyglet.window.key, i), int):
				l.append(i)
		return l

	@configurable
	def get_key(self):

		keylist = self.keylist
		if keylist is not None:
			keylist += [u'escape']
		timeout = self.timeout
		start_time = self.experiment.clock.time()
		while True:
			# Don't use the timeStamped keyword, because it bugs on older
			# versions of PsychoPy (at least 1.79.01 under Ubuntu 14.04).
			# See https://github.com/smathot/OpenSesame/issues/381
			time = self.experiment.clock.time()
			keys = event.getKeys(keylist, modifiers=True)
			for key, modifiers in keys:
				if key == u'escape':
					self.experiment.pause()
				if keylist is None or key in keylist:
					self._latest_modifiers = modifiers
					return key, time
			if timeout is not None and time - start_time >= timeout:
				break
		self._latest_modifiers = {}
		return None, time

	def get_mods(self):

		return [
			m
			for m in (u'shift', u'ctrl', u'alt')
			if self._latest_modifiers.get(m, False)
		]

	def synonyms(self, key):

		if key is None:
			return [None, 'None', 'none', 'NONE']

		# Respond correctly if a keycode is passed, rather than a Unicode string
		# key description.
		if type(key) == int:
			l = [key, pyglet.window.key.symbol_string(key).lower()]
			if l[-1].upper() != l[-1].lower():
				l.append(l[-1].upper())
			return l

		# Sanity check
		if not isinstance(key, basestring):
			raise osexception(
				u'Key names should be string or numeric, not %s' % type(key))

		# Make a list of all conceivable ways that a key might be referred to.
		l = [key]
		if key != key.upper():
			l.append(key.upper())
		if key != key.lower():
			l.append(key.lower())
		for char, name in self.keymap.items():
			if key == char:
				l.append(name)
			if key.lower() == name:
				l.append(char)
		# Make sure that we can deal with None/ timeout responses
		if key.lower() == u'none':
			l.append(None)
		# Make sure that we convert numeric strings to ints as well
		try:
			i = int(key)
			l.append(i)
		except:
			pass
		return l

	def flush(self):

		keypressed = False
		for key in event.getKeys():
			if key == u"escape":
				self.experiment.pause()
			keypressed = True
		return keypressed

	def _keycode_to_str(self, keycode):

		return pyglet.window.key.symbol_string(keycode).lower()


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import json
from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libqtopensesame.extensions import base_extension
from libopensesame import misc
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'after_experiment', category=u'extension')


class after_experiment(base_extension):

	"""
	desc:
		Shows notifications after an experiment has finished.
	"""

	def event_end_experiment(self, ret_val):

		"""
		desc:
			Handles the end of an experiment.

		arguments:
			ret_val:
				desc:	An Exception, or None if no exception occurred.
				type:	[Exception, NoneType]
		"""

		if ret_val is None:
			self.handle_success()
		else:
			self.handle_exception(ret_val)

	@property
	def _logfile(self):

		"""
		returns:
			desc:	The path to the logfile of the last run, or None if this
					could not be determined.
			type:	[str, NoneType]
		"""

		# Depending on the runner, the logfile is either a property of the
		# var object ...
		var = self.extension_manager.provide(
			'jupyter_workspace_variable',
			name='var'
		)
		try:
			if var and 'logfile' in var:
				return var.logfile
		except TypeError:
			# JupyterConsole returns a custom FailedToGetWorkspaceVariable
			# Exception when an error occurs. This is not iterable and results
			# in a TypeError.
			pass
		# ... or it is a global variable in the workspace
		logfile = self.extension_manager.provide(
			'jupyter_workspace_variable',
			name='logfile'
		)
		if logfile:
			return logfile

	@property
	def _extra_data_files(self):

		"""
		returns:
			A list of extra data files, such as eye-tracking data.
		"""

		data_files = self.extension_manager.provide(
			'jupyter_workspace_variable',
			name='data_files'
		)
		if not data_files:
			return
		return [
			path for path in data_files
			if path != self._logfile
		]

	def event_after_experiment_copy_logfile(self):

		"""
		desc:
			Copies the logfile to the file pool.
		"""

		if self._logfile is None:
			return
		self.main_window.ui.pool_widget.add([self._logfile],
 			rename=True)

	def event_after_experiment_open_logfile_folder(self):

		"""
		desc:
			Opens the logfile folder.
		"""

		if self._logfile is None:
			return
		misc.open_url(os.path.dirname(self._logfile))

	def event_after_experiment_open_logfile(self):

		"""
		desc:
			Opens the logfile.
		"""

		if self._logfile is None:
			return
		misc.open_url(self._logfile)

	def handle_success(self):

		"""
		desc:
			Shows a summary after successful completion of the experiment.
		"""

		logfile = self._logfile
		if logfile is None:
			logfile = u'Unknown logfile'
		md = safe_read(self.ext_resource(u'finished.md')) % {
			u'time': time.ctime(),
			u'logfile': logfile
			}
		if self._extra_data_files:
			print(self._extra_data_files)
			md += u'\n' \
				+ _(u'The following extra data files where created:') + u'\n\n'
			for data_file in self._extra_data_files:
				md += u'- `' + data_file + u'`\n'
			md += u'\n'
		self.tabwidget.open_markdown(md, u'os-finished-success', _(u'Finished'))

	def handle_exception(self, e):

		"""
		desc:
			Shows a summary when the experiment was aborted.

		arguments:
			e:
				desc:	The Exception that caused the experiment to stop.
				type:	Exception
		"""

		if not isinstance(e, osexception):
			e = osexception(msg=u'Unexpected error', exception=e)
		if e.user_triggered:
			icon = u'os-finished-user-interrupt'
			title = _(u'Aborted')
			md = _(u'# Aborted\n\n- ') + e.markdown()
		else:
			icon = u'os-finished-error'
			title = _(u'Stopped')
			md = _(u'# Stopped\n\nThe experiment did not finish normally for the following reason:\n\n- ') \
				+ e.markdown()
		self.console.write(e)
		self.tabwidget.open_markdown(md, icon, title)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from qtpy import QtCore, QtWidgets
from libqtopensesame.dialogs.base_dialog import base_dialog
from libqtopensesame.widgets import sketchpad_widget

class sketchpad_dialog(base_dialog):

	"""
	desc:
		The pop-out version of the sketchpad_widget.
	"""

	def __init__(self, main_window, sketchpad):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	A qtopensesame object.
			sketchpad:		A sketchpad object.
		"""

		super(sketchpad_dialog, self).__init__(main_window, flags= \
			QtCore.Qt.WindowMinMaxButtonsHint | QtCore.Qt.WindowCloseButtonHint)
		self.sketchpad = sketchpad
		self.tools_widget = sketchpad_widget.sketchpad_widget(self.sketchpad,
			parent=self, embed=False)
		self.close_button = QtWidgets.QPushButton(self.theme.qicon(u"close"),
			u"Close")
		self.close_button.setIconSize(QtCore.QSize(16,16))
		self.close_button.clicked.connect(self.accept)
		self.hbox = QtWidgets.QHBoxLayout()
		self.hbox.addStretch()
		self.hbox.addWidget(self.close_button)
		self.hbox.setContentsMargins(0, 0, 0, 0)
		self.hbox_widget = QtWidgets.QWidget()
		self.hbox_widget.setLayout(self.hbox)
		self.vbox = QtWidgets.QVBoxLayout()
		self.vbox.addWidget(self.tools_widget)
		self.vbox.addWidget(self.hbox_widget)
		self.setLayout(self.vbox)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import debug
from libqtopensesame.widgets.header_widget import header_widget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'general_header_widget', category=u'core')

class general_header_widget(header_widget):

	"""
	desc:
		The widget containing the clickable title and description of the
		experiment.
	"""

	def __init__(self, general_tab, main_window):

		"""
		desc:
			Constructor.

		arguments:
			general_tab:	A general_properties object.
			main_window:	A qtopensesame object.
		"""

		header_widget.__init__(self, main_window.experiment)
		self.general_tab = general_tab

	def refresh(self):

		"""
		desc:
			Updates the header so that it's content match the item.
		"""

		self.set_name(self.experiment.var.title)
		self.set_desc(self.experiment.var.description)

	def apply_name(self):

		"""
		desc:
			Applies the name change and revert the edit control back to the
			static label.
		"""

		if self.label_name.isVisible():
			return
		self.label_name.show()
		self.label_type.show()
		self.edit_name.hide()
		self.general_tab.apply_changes()
		self.refresh()

	def apply_desc(self):

		"""
		desc:
			Applies the description change and revert the edit back to the
			label.
		"""

		if self.label_desc.isVisible():
			return
		self.label_desc.show()
		self.edit_desc.hide()
		self.general_tab.apply_changes()
		self.refresh()



#-*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.

---
desc:
	Drag-and-drop data is communicated by json-encoded dictionaries. The
	following drag types exist:

	-	type:			invalid

	-	type:			sketchpad-element-move

	-	type:			url-local
		url:			[url]

	-	type:			item-existing
		item-name:		[item name]
		QTreeWidgetItem	[representation of QTreeWidgetItem]

	-	type:		item-clone
		item:		[item name]

	-	type:		variable
		variable:	[variable name]

	-	type:			item-snippet
		main-item-name:	[suggested name]

		items:
			-	item-name:	[suggested name]
				item-type:	[item type]
				script:		[script]
			-	[...]
---
"""

import json
from qtpy import QtCore, QtGui
from libopensesame.oslogging import oslogger
from libopensesame.py3compat import *


# For unknown mimetypes
INVALID_DATA = {
	u'type' : u'invalid'
}


def matches(data, types):

	"""
	desc:
		Checks whether a data dictionary matches any of the specified drop
		types.

	arguments:
		data:
			desc:	A data dictionary. Non-dict types do not give an error, but
					return False.
		types:
			desc:	A list of types, matching the 'type' key in the data dict.
			type:	list

	returns:
		desc:		True if the data matches, False otherwise.
		type:		bool
	"""

	if not isinstance(data, dict):
		return False
	if u'type' not in data:
		return False
	if data[u'type'] not in types:
		return False
	return True

def receive(drop_event):

	"""
	desc:
		Extracts data from a drop event. This data should be embedded in the
		mimedata as a json text string.

	arguments:
		drop_event:
			desc:	A drop event.
			type:	QDropEvent

	returns:
		desc:		A data dictionary. The 'type' key identifies the type of
					data that is being dropped.
		type:		dict
	"""

	mimedata = drop_event.mimeData()
	# Reject unknown mimedata
	if not mimedata.hasText() and not mimedata.hasUrls():
		oslogger.warning(u'No text data')
		return INVALID_DATA
	# Process url mimedata
	if mimedata.hasUrls():
		urls = mimedata.urls()
		if len(urls) == 0:
			return INVALID_DATA
		url = urls[0]
		if not url.isLocalFile():
			return INVALID_DATA
		path = str(url.toLocalFile())
		data = {
			u'type': u'url-local',
			u'url': path
		}
		return data
	# Process JSON text data
	text = str(mimedata.text())
	try:
		data = json.loads(text)
	except:
		oslogger.warning(u'Failed to load json mime data')
		return INVALID_DATA
	if not isinstance(data, dict) or u'type' not in data:
		return INVALID_DATA
	return data

def send(drag_src, data):

	"""
	desc:
		Starts a drag event, and embeds a data dictionary as json text in the
		mimedata.

	arguments:
		drag_src:
			desc:	The source widget.
			type:	QWidget
		data:
			desc:	A data dictionary. The 'type' key identifies the type of
					data that is being dropped.
			type:	dict

	returns:
		desc:		A drag object. The start function is called automatically.
		type:		QDrag
	"""

	text = safe_decode(json.dumps(data), enc=u'utf-8')
	mimedata = QtCore.QMimeData()
	mimedata.setText(text)
	drag = QtGui.QDrag(drag_src)
	drag.setMimeData(mimedata)
	drag.exec_()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

---
desc:
	Functions that are globally accessible in `inline_script` items.
---
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
import random
import math
import warnings
# The classes below are unused, but imported so that they are available in the
# workspace.
from openexp.canvas_elements import (Rect, Line, Text, Ellipse, Circle,
	FixDot, Gabor, NoisePatch, Image, Arrow, Polygon)
from libopensesame.widgets.widget_factory import (Label, Button, ImageWidget,
	ImageButton, TextInput, RatingScale, Checkbox)


# Factory functions


def Form(*args, **kwargs):

	"""
	desc: |
		A factory function that creates a new `Form` object. For a
		description of possible keywords, see:

		- %link:manual/forms/widgets%

	returns:
		desc:	A `Form` object.
		type:	canvas

	example: |
		form = Form()
		label = Label(text='label')
		button = Button(text='Ok')
		form.set_widget(label, (0,0))
		form.set_widget(button, (0,1))
		form._exec()
	"""

	from libopensesame.widgets import form
	return form(experiment, **kwargs)


def Canvas(auto_prepare=True, **style_args):

	"""
	desc: |
		A factory function that creates a new `Canvas` object. For a
		description of possible keywords, see:

		- %link:manual/python/canvas%

	returns:
		desc:	A `Canvas` object.
		type:	canvas

	example: |
		my_canvas = Canvas(color=u'red', penwidth=2)
		my_canvas.line(-10, -10, 10, 10)
		my_canvas.line(-10, 10, 10, -10)
		my_canvas.show()
	"""

	from openexp.canvas import Canvas
	return Canvas(experiment, auto_prepare=auto_prepare, **style_args)


def Keyboard(**resp_args):

	"""
	desc: |
		A factory function that creates a new `Keyboard` object. For a
		description of possible keywords, see:

		- %link:manual/python/keyboard%

	returns:
		desc:	A `Keyboard` object.
		type:	keyboard

	example: |
		my_keyboard = Keyboard(keylist=[u'a', u'b'], timeout=5000)
		key, time = my_keyboard.get_key()
	"""

	from openexp.keyboard import Keyboard
	return Keyboard(experiment, **resp_args)


def Mouse(**resp_args):

	"""
	desc: |
		A factory function that creates a new `Mouse` object. For a
		description of possible keywords, see:

		- %link:manual/python/mouse%

	returns:
		desc:	A `mouse` object.
		type:	mouse

	example: |
		my_mouse = Mouse(keylist=[1,3], timeout=5000)
		button, time = my_mouse.get_button()
	"""

	from openexp.mouse import Mouse
	return Mouse(experiment, **resp_args)


def Sampler(src, **playback_args):

	"""
	desc: |
		A factory function that creates a new `Sampler` object. For a
		description of possible keywords, see:

		- %link:manual/python/sampler%

	returns:
		desc:	A SAMPLER object.
		type:	sampler

	example: |
		src = pool['bark.ogg']
		my_sampler = Sampler(src, volume=.5, pan='left')
		my_sampler.play()
	"""

	from openexp.sampler import Sampler
	return Sampler(experiment, src, **playback_args)


# Miscellaneous API	functions


def Synth(osc="sine", freq=440, length=100, attack=0, decay=5):

	"""
	desc:
		A factory function that synthesizes a sound and returns it as a
		`Sampler` object.

	keywords:
		osc:
			desc:	Oscillator, can be "sine", "saw", "square" or
					"white_noise".
			type:	[str, unicode]
		freq:
			desc:	Frequency, either an integer value (value in hertz) or a
					string ("A1", "eb2", etc.).
			type:	[str, unicode, int, float]
		length:
			desc:	The length of the sound in milliseconds.
			type:	[int, float]
		attack:
			desc:	The attack (fade-in) time in milliseconds.
			type:	[int, float]
		decay:
			desc:	The decay (fade-out) time in milliseconds.
			type:	[int, float]

	returns:
		desc:	A SAMPLER object.
		type:	sampler

	example: |
		my_sampler = Synth(freq=u'b2', length=500)
	"""

	from openexp.synth import Synth
	return Synth(experiment, osc=osc, freq=freq, length=length, attack=attack,
		decay=decay)


def copy_sketchpad(name):

	"""
	desc:
		Returns a copy of a `sketchpad`'s canvas.

	arguments:
		name:
			desc:	The name of the `sketchpad`.
			type:	[str, unicode]

	returns:
		desc:	A copy of the `sketchpad`'s canvas.
		type:	canvas

	example: |
		my_canvas = copy_sketchpad('my_sketchpad')
		my_canvas.show()
	"""

	c = Canvas()
	c.copy(experiment.items[name].canvas)
	return c


def reset_feedback():

	"""
	desc:
		Resets all feedback variables to their initial state.

	example: |
		reset_feedback()
	"""

	experiment.reset_feedback()


def set_response(response=None, response_time=None, correct=None):

	"""
	visible: False
	"""

	warnings.warn(
		u'set_response() has been deprecated. Use responses object instead.',
		DeprecationWarning)
	experiment.responses.add(response=response, response_time=response_time,
		correct=correct)

def set_subject_nr(nr):

	"""
	desc:
		Sets the subject number and parity (even/ odd). This function is called
		automatically when an experiment is started, so you only need to call it
		yourself if you overwrite the subject number that was specified when the
		experiment was launched.

	arguments:
		nr:
			desc:	The subject nr.
			type:	int

	example: |
		set_subject_nr(1)
		print('Subject nr = %d' % var.subject_nr)
		print('Subject parity = %s' % var.subject_parity)
	"""

	experiment.set_subject(nr)


def sometimes(p=.5):

	"""
	desc: |
		Returns True with a certain probability. (For more advanced
		randomization, use the Python `random` module.)

	keywords:
		p:
			desc:	The probability of returning True.
			type:	float

	returns:
		desc:	True or False
		type:	bool

	example: |
		if sometimes():
			print('Sometimes you win')
		else:
			print('Sometimes you loose')
	"""

	if (not isinstance(p, float) and not isinstance(p, int)) or p < 0 or p > 1:
		raise osexception(
			u'p should be a numeric value between 0 and 1, not "%s"' % p)
	return random.random() < p


def pause():

	"""
	desc:
		Pauses the experiment.
	"""

	experiment.pause()


def register_cleanup_function(fnc):

	"""
	desc:
		Registers a clean-up function, which is executed when the experiment
		ends. Clean-up functions are executed at the very end, after the
		display, sound device, and log file have been closed. Clean-up functions
		are also executed when the experiment crashes.

	argument:
		fnc:
			desc:	A function that doesn't require arguments.
			type:	callable

	example: |
		def my_cleanup_function():
			print(u'The experiment is finished!')
		register_cleanup_function(my_cleanup_function)
	"""

	experiment.cleanup_functions.append(fnc)


def xy_from_polar(rho, phi, pole=(0, 0)):

	"""
	desc:
		Converts polar coordinates (distance, angle) to Cartesian coordinates
		(x, y).

	arguments:
		rho:
			desc:	The radial coordinate, also distance or eccentricity.
			type:	float
		phi:
			desc:	The angular coordinate. This reflects a clockwise rotation
					in degrees (i.e. not radians), where 0 is straight
					right.
			type:	float

	keywords:
		pole:
			desc:	The refence point.
			type:	tuple

	returns:
		desc:	An (x, y) coordinate tuple.
		type:	tuple

	example: |
		# Draw a cross
		x1, y1 = xy_from_polar(100, 45)
		x2, y2 = xy_from_polar(100, -45)
		c = Canvas()
		c.line(x1, y1, -x1, -y1)
		c.line(x2, y2, -x2, -y2)
		c.show()
	"""

	try:
		rho = float(rho)
	except:
		raise osexception(u'rho should be numeric in xy_from_polar()')
	try:
		phi = float(phi)
	except:
		raise osexception(u'phi should be numeric in xy_from_polar()')
	phi = math.radians(phi)
	ox, oy = parse_pole(pole)
	x = rho * math.cos(phi) + ox
	y = rho * math.sin(phi) + oy
	return x, y


def xy_to_polar(x, y, pole=(0, 0)):

	"""
	desc:
		Converts Cartesian coordinates (x, y) to polar coordinates (distance,
		angle).

	arguments:
		x:
			desc:	The X coordinate.
			type:	float
		y:
			desc:	The Y coordinate.
			type:	float

	keywords:
		pole:
			desc:	The refence point.
			type:	tuple

	returns:
		desc:	An (rho, phi) coordinate tuple. Here, `rho` is the radial
				coordinate, also distance or eccentricity. `phi` is the angular
				coordinate in degrees (i.e. not radians), and reflects a
				counterclockwise rotation, where 0 is straight right.
		type:	tuple

	example: |
		rho, phi = xy_to_polar(100, 100)
	"""

	try:
		x = float(x)
	except:
		raise osexception(u'x should be numeric in xy_to_polar()')
	try:
		y = float(y)
	except:
		raise osexception(u'y should be numeric in xy_to_polar()')
	ox, oy = parse_pole(pole)
	dx = x-ox
	dy = y-oy
	rho = math.sqrt(dx**2 + dy**2)
	phi = math.degrees(math.atan2(dy, dx))
	return rho, phi


def xy_distance(x1, y1, x2, y2):

	"""
	desc:
		Gives the distance between two points.

	arguments:
		x1:
			desc:	The x coordinate of the first point.
			type:	float
		y1:
			desc:	The y coordinate of the first point.
			type:	float
		x2:
			desc:	The x coordinate of the second point.
			type:	float
		y2:
			desc:	The y coordinate of the second point.
			type:	float

	returns:
		desc:	The distance between the two points.
		type:	float
	"""

	try:
		x1 = float(x1)
		y1 = float(y1)
		x2 = float(x2)
		y2 = float(y2)
	except:
		raise osexception(u'Coordinates should be numeric in xy_distance()')
	return math.sqrt((x1-x2)**2+(y1-y2)**2)


def xy_circle(n, rho, phi0=0, pole=(0, 0)):

	"""
	desc:
		Generates a list of points (x,y coordinates) in a circle. This can be
		used to draw stimuli in a circular arrangement.

	arguments:
		n:
			desc:	The number of x,y coordinates to generate.
			type:	int
		rho:
			desc:	The radial coordinate, also distance or eccentricity, of the
					first point.
			type:	float

	keywords:
		phi0:
			desc:	The angular coordinate for the first coordinate. This is a
					counterclockwise rotation in degrees (i.e. not radians),
					where 0 is straight right.
			type:	float
		pole:
			desc:	The refence point.
			type:	tuple

	returns:
		desc:	A list of (x,y) coordinate tuples.
		type:	list

	example: |
		# Draw 8 rectangles around a central fixation dot
		c = Canvas()
		c.fixdot()
		for x, y in xy_circle(8, 100):
			c.rect(x-10, y-10, 20, 20)
		c.show()
	"""

	if not isinstance(n, int) or n < 0:
		raise osexception(u'n should be a non-negative integer in xy_circle()')
	try:
		phi0 = float(phi0)
	except:
		raise osexception(u'phi0 should be numeric in xy_circle()')
	l = []
	for i in range(n):
		l.append(xy_from_polar(rho, phi0, pole=pole))
		phi0 += 360./n
	return l


def xy_grid(n, spacing, pole=(0, 0)):

	"""
	desc:
		Generates a list of points (x,y coordinates) in a grid. This can be
		used to draw stimuli in a grid arrangement.

	arguments:
		n:
			desc:	An `int` that indicates the number of columns and rows, so
					that `n=2` indicates a 2x2 grid, or a (n_col, n_row)
					`tuple`, so that `n=(2,3)` indicates a 2x3 grid.
			type:	[int, tuple]
		spacing:
			desc:	A numeric value that indicates the spacing between cells, or
					a (col_spacing, row_spacing) tuple.
			type:	float

	keywords:
		pole:
			desc:	The refence point.
			type:	tuple

	returns:
		desc:	A list of (x,y) coordinate tuples.
		type:	list

	example: |
		# Draw a 4x4 grid of rectangles
		c = Canvas()
		c.fixdot()
		for x, y in xy_grid(4, 100):
			c.rect(x-10, y-10, 20, 20)
		c.show()
	"""

	try:
		n_col, n_row = n
	except:
		n_col = n_row = n
	try:
		n_col = int(n_col)
		n_row = int(n_row)
		assert(n_col >= 0)
		assert(n_row >= 0)
	except:
		raise osexception(u'n should be a non-negative integer or a tuple of '
				u'two non-negative integers in xy_grid()')
	try:
		s_col, s_row = spacing
	except:
		s_col = s_row = spacing
	try:
		s_col = float(s_col)
		s_row = float(s_row)
		assert(s_col >= 0)
		assert(s_row >= 0)
	except:
		raise osexception(u'spacing should be a non-negative numeric or a '
			u'tuple of two non-negative numerics in xy_grid()')
	pole = parse_pole(pole)
	l = []
	for row in range(n_row):
		y = (row - (n_row-1) / 2.) * s_row + pole[1]
		for col in range(n_col):
			x = (col - (n_col-1) / 2.) * s_col + pole[0]
			l.append((x, y))
	return l


def xy_random(n, width, height, min_dist=0, pole=(0,0)):

	"""
	desc:
		Generates a list of random points (x,y coordinates) with a minimum
		spacing between each pair of points. This function will raise an
		`osexception` when the coordinate list cannot be generated, typically
		because there are too many points, the min_dist is set too high, or the
		width or height are set too low.

	arguments:
		n:
			desc:	The number of points to generate.
			type:	int
		width:
			desc:	The width of the field with random points.
			type:	float
		height:
			desc:	The height of the field with random points.
			type:	float

	keywords:
		min_dist:
			desc:	The minimum distance between each point.
			type:	float
		pole:
			desc:	The refence point.
			type:	tuple

	returns:
		desc:	A list of (x,y) coordinate tuples.
		type:	list

	example: |
		# Draw a 50 rectangles in a random grid
		c = Canvas()
		c.fixdot()
		for x, y in xy_random(50, 500, 500, min_dist=40):
			c.rect(x-10, y-10, 20, 20)
		c.show()
	"""

	if not isinstance(n, int) or n < 0:
		raise osexception(u'n should be a non-negative integer in xy_random()')
	try:
		width = float(width)
	except:
		raise osexception(u'width should be numeric in xy_random()')
	try:
		height = float(height)
	except:
		raise osexception(u'height should be numeric in xy_random()')
	try:
		min_dist = float(min_dist)
	except:
		raise osexception(u'min_dist should be numeric in xy_random()')
	pole = parse_pole(pole)
	max_try = 1000
	for t1 in range(max_try):
		l = []
		for i in range(n):
			for t2 in range(max_try):
				x1 = (random.random()-.5)*width + pole[0]
				y1 = (random.random()-.5)*height + pole[1]
				for x2, y2 in l:
					if xy_distance(x1, y1, x2, y2) < min_dist:
						break
				else:
					# Point does not collide, so add to the list
					l.append( (x1, y1) )
					break
			else:
				# All level-2 tries have failed, so break to start a new level-1
				# try.
				break
		else:
			# All points have been successfully added, so return the list
			return l
	# All level-1 tries have failed
	raise osexception(
		u'Failed to generate random coordinates in xy_random()')


# Helper functions that are not part of the public API


def parse_pole(pole):

	"""
	visible: False
	"""

	try:
		ox = float(pole[0])
		oy = float(pole[1])
		assert(len(pole) == 2)
	except:
		raise osexception(u'pole should be a tuple (or similar) of length '
			u'with two numeric values')
	return ox, oy


def set_aliases():

	"""
	visible: False
	"""

	# Non PEP-8 alias for backwards compatibility
	global canvas, sampler, synth, keyboard, mouse
	canvas = Canvas
	sampler = Sampler
	synth = Synth
	keyboard = Keyboard
	mouse = Mouse



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._element.element import Element


class Line(Element):

	CONTAINS_MAX_DIST = 10

	def __init__(self, canvas, sx, sy, ex, ey, **properties):

		properties = properties.copy()
		properties.update({'sx' : sx, 'sy' : sy, 'ex' : ex, 'ey' : ey})
		self._shapely_line = None
		Element.__init__(self, canvas, **properties)

	@property
	def rect(self):

		return self._rect(self.sx, self.sy, self.ex-self.sx, self.ey-self.sy)

	def __contains__(self, xy):

		# Shapely is used to determine whether a point is close enough to the
		# line to be counted as 'inside' it. If shapely isn't available, we fall
		# back to a simple bounding box. The shapely LineString is stored for
		# performance.
		try:
			from shapely.geometry import LineString, Point
		except ImportError:
			return Element.__contains__(self, xy)
		if self._shapely_line is None:
			self._shapely_line = LineString([
				(self.sx, self.sy),
				(self.ex, self.ey)
			])
		return Point(*xy).distance(self._shapely_line) < self.CONTAINS_MAX_DIST

	def _on_attribute_change(self, **kwargs):

		# When a change occurs the shapely line (if any) needs to be
		# redetermined in __contains__
		self._shapely_line = None
		Element._on_attribute_change(self, **kwargs)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import re
import types
from openexp.canvas_elements import ElementFactory
from libopensesame import python_workspace_api as api
from libopensesame.widgets.widget_factory import WidgetFactory
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import base_extension
from pyqode.core import icons
from pyqode.core.widgets import SplittableCodeEditTabWidget
from pyqode_extras.widgets import (
	PythonCodeEdit,
	FallbackCodeEdit
)
from pyqode.core.api import ColorScheme
from pyqode.core.api.panel import Panel
from pyqode.core.panels import LineNumberPanel, FoldingPanel, MarginPanel
from pyqode.core.modes import (
	PygmentsSH,
	RightMarginMode,
	CodeCompletionMode,
	AutoCompleteMode,
	BreakpointMode
)
from pyqode.python.modes import (
	sh,
	PyAutoCompleteMode,
	CalltipsMode
)
from pyqode.core.managers import backend
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import QPlainTextEdit


class pyqode_manager(base_extension):

	"""
	desc:
		Keeps track of all PyQode editor instances, which need to be explicitly
		closed. And also modifies their behavior somewhat.
	"""

	def _is_opensesame_api_object(self, name, obj):

		if name.startswith(u'safe_'):
			return False
		if isinstance(obj, types.FunctionType):
			return True
		if isinstance(obj, type) and issubclass(
			obj, (WidgetFactory, ElementFactory)):
				return True
		return False

	def _register_opensesame_builtins(self):

		builtins = [
			u'exp',
			u'var',
			u'pool',
			u'items',
			u'clock',
			u'log',
			u'responses',
			u'data_files',
			u'AbortCoroutines'
		] + [
			name
			for name, obj in api.__dict__.items()
			if self._is_opensesame_api_object(name, obj)
		]
		sh.PythonSH.PROG = re.compile(
			sh.make_python_patterns(additional_builtins=builtins),
			re.S
		)
		
	def settings_widget(self):
		
		from pyqode_preferences import PyQodePreferences
		return PyQodePreferences(self.main_window)

	def event_startup(self):

		backend.comm = oslogger.debug
		if self.main_window.mode == u'default':
			self._register_opensesame_builtins()
		icons.USE_QTAWESOME = True
		icons.QTA_OPTIONS = {
			'color': self.main_window.palette().text().color().name(),
			'color_disabled': self.main_window.palette().mid().color().name(),
			'color_info': '#66d9ef',
			'color_warning': '#e6db74',
			'color_error': '#f92672'
		}
		SplittableCodeEditTabWidget.editors = {}
		SplittableCodeEditTabWidget.fallback_editor = FallbackCodeEdit
		SplittableCodeEditTabWidget.register_code_edit(PythonCodeEdit)
		SplittableCodeEditTabWidget.register_code_edit(FallbackCodeEdit)
		# Syntax is broken for TextCodeEdit, so we don't use it for now
		# SplittableCodeEditTabWidget.register_code_edit(TextCodeEdit)
		self._editors = []
		self._breakpoints = {}
		self._auto_backend_restart = True

	def event_close(self):

		self._auto_backend_restart = False
		oslogger.debug(u'closing ({})'.format(len(self._editors)))
		while self._editors:
			self._editors.pop().close()
		self._auto_backend_restart = True

	def event_prepare_delete_item(self, name):

		item = self.item_store[name]
		if item.container_widget is None:
			return
		if item.item_type == u'inline_script':
			self.event_unregister_editor(item._pyqode_run_editor)
			self.event_unregister_editor(item._pyqode_prepare_editor)
		if hasattr(self, u'auto_editor'):
			for editor in self.auto_editor.values():
				self.event_unregister_editor(editor)
		self.event_unregister_editor(item._script_widget)

	def event_run_experiment(self, fullscreen):

		self._auto_backend_restart = False
		for editor in self._editors:
			editor.backend.suspend()
		self._auto_backend_restart = True

	def event_end_experiment(self, ret_val):

		for editor in self._editors:
			editor.backend.resume()
			self._register_backend_process(editor)

	def event_prepare_purge_unused_items(self):

		for name in self.item_store.unused():
			self.event_prepare_delete_item(name)

	def event_open_experiment(self, path):

		self.event_close()

	def event_prepare_regenerate(self):

		self.event_close()

	def event_register_editor(self, editor, mime_type=u'text/plain'):

		editor._auto_reset_stylesheet = False
		self._editors.append(editor)		
		if 'PythonSH' in editor.modes.keys():
			pass
		elif 'TextSH' in editor.modes.keys():
			editor.modes.get('TextSH').color_scheme = safe_str(
				cfg.pyqode_color_scheme
			)
		elif 'PygmentsSH' in editor.modes.keys():
			editor.modes.get('PygmentsSH').color_scheme = safe_str(
				cfg.pyqode_color_scheme
			)
		else:
			sh = PygmentsSH(
				editor.document(),
				color_scheme=ColorScheme('monokai')
			)
			sh.set_mime_type(mime_type)
			editor.modes.append(sh)
		if cfg.pyqode_enable_breakpoints:
			editor.modes.append(BreakpointMode(breakpoints=self._breakpoints))
		editor.font_size = cfg.pyqode_font_size
		editor.font_name = cfg.pyqode_font_name
		editor.setLineWrapMode(
			QPlainTextEdit.WidgetWidth
			if cfg.pyqode_line_wrap
			else QPlainTextEdit.NoWrap
		)
		editor.action_duplicate_line.setShortcut(
			cfg.pyqode_shortcut_duplicate_line
		)
		editor.action_swap_line_up.setShortcut(
			cfg.pyqode_shortcut_swap_line_up
		)
		editor.action_swap_line_down.setShortcut(
			cfg.pyqode_shortcut_swap_line_down
		)
		editor._auto_reset_stylesheet = True
		editor._reset_stylesheet()
		editor.setTabStopWidth(
			QFontMetrics(
				editor.font()
			).averageCharWidth() * cfg.pyqode_tab_length
		)
		# This function is defined here so to keep the editor in the scope.
		# This is necessary in case the backend crashes and it needs to be
		# restarted.
		def auto_restart():
			if editor._closed or not self._auto_backend_restart:
				return
			oslogger.warning(
				u'restarting backend: {}'.format(editor.__class__.__name__)
			)
			if editor.backend._share_id in editor.backend.SHARE_COUNT:
				del editor.backend.SHARE_COUNT[editor.backend._share_id]
			editor.backend._process = None
			editor.backend.resume()
			self._register_backend_process(editor)
		try:
			editor.backend._process.error.connect(auto_restart)
			self._register_backend_process(editor)
		except RuntimeError:
			auto_restart()
		oslogger.debug(u'registering {}'.format(editor))

	def event_unregister_editor(self, editor):

		if editor is None:
			return
		self._auto_backend_restart = False
		editor.close()
		if editor in self._editors:
			self._editors.remove(editor)
		oslogger.debug(u'unregistering ({})'.format(len(self._editors)))
		self._auto_backend_restart = True
		
	def event_setting_changed(self, setting, value):
		
		fnc = '_setting_{}'.format(setting)
		if hasattr(self, fnc):
			getattr(self, fnc)(value)
			
	def provide_pyqode_breakpoints(self):
		
		return self._breakpoints

	def event_pyqode_suspend_auto_backend_restart(self):
		
		self._auto_backend_restart = False

	def event_pyqode_resume_auto_backend_restart(self):
		
		self._auto_backend_restart = True

	def event_pyqode_set_breakpoints(self, breakpoints):
		
		self._breakpoints = breakpoints

	def event_pyqode_clear_breakpoints(self):
		
		self.extension_manager.fire(
			'notify',
			message=_(u'Cleared {} breakpoint(s) in {} file(s)'.format(
				sum(len(lines) for lines in self._breakpoints.values()),
				len(self._breakpoints))
			)
		)
		self._breakpoints.clear()
		for editor in self._editors:
			if 'BreakpointMode' not in editor.modes.keys():
				continue
			editor.modes.get('BreakpointMode').clear_messages()
			editor.repaint()
			
	def _register_backend_process(self, editor):
		
		self.extension_manager.fire(
			'register_subprocess',
			pid=editor.backend._process.processId(),
			description='BackendProcess:{}'.format(
				editor.__class__.__name__
			)
		)

	def _setting_pyqode_show_line_numbers(self, show_line_numbers):

		self._toggle_panel(
			'pyqode_show_line_numbers',
			show_line_numbers,
			LineNumberPanel,
			Panel.Position.LEFT
		)

	def _setting_pyqode_line_wrap(self, line_wrap):

		cfg.pyqode_line_wrap = line_wrap
		for editor in self._editors:
			editor.setLineWrapMode(
				QPlainTextEdit.WidgetWidth
				if cfg.pyqode_line_wrap
				else QPlainTextEdit.NoWrap
			)

	def _setting_pyqode_code_completion(self, code_completion):

		editors = self._editors
		self._editors = [
			e for e in editors
			if isinstance(e, FallbackCodeEdit) and
			e.language not in cfg.pyqode_code_completion_excluded_mimetypes
		]
		for mode in (
			CodeCompletionMode,
			AutoCompleteMode
		):
			self._toggle_mode('pyqode_code_completion', code_completion, mode)
		self._editors = [
			e for e in editors
			if isinstance(e, PythonCodeEdit) and
			e.language not in cfg.pyqode_code_completion_excluded_mimetypes
		]
		for mode in (
			CodeCompletionMode,
			PyAutoCompleteMode,
			CalltipsMode
		):
			self._toggle_mode('pyqode_code_completion', code_completion, mode)
		self._editors = editors

	def _setting_pyqode_fixed_width(self, fixed_width):

		self._toggle_panel(
			'pyqode_fixed_width',
			fixed_width,
			MarginPanel,
			Panel.Position.LEFT,
			nchar=cfg.pyqode_fixed_width_nchar
		)

	def _setting_pyqode_show_whitespaces(self, show_whitespaces):

		cfg.pyqode_show_whitespaces = show_whitespaces
		for editor in self._editors:
			editor.show_whitespaces = cfg.pyqode_show_whitespaces

	def _setting_pyqode_right_margin(self, show_right_margin):

		self._toggle_mode(
			'pyqode_right_margin',
			show_right_margin,
			RightMarginMode
		)

	def _setting_pyqode_code_folding(self, code_folding):

		self._toggle_panel(
			'pyqode_code_folding',
			code_folding,
			FoldingPanel,
			Panel.Position.LEFT
		)

	def event_pyqode_select_indentation_mode(self):

		def _select_indentation_mode(mode):

			cfg.pyqode_indentation = mode
			for editor in self._editors:
				if 'AutodetectIndentationMode' not in editor.modes.keys():
					continue
				editor.modes.get('AutodetectIndentationMode')._autodetect_indentation()

		haystack = [
			(_('autodetect'), 'autodetect', _select_indentation_mode),
			(_('spaces'), 'spaces', _select_indentation_mode),
			(_('tabs'), 'tabs', _select_indentation_mode)
		]
		self.extension_manager.fire(
			u'quick_select',
			haystack=haystack,
			placeholder_text=_(u'Select indentation mode …')
		)

	def _toggle_mode(self, config, enabled, cls, **kwargs):

		cfg[config] = enabled
		for editor in self._editors:
			if enabled:
				if cls.__name__ in editor.modes.keys():
					continue
				editor.modes.append(cls(**kwargs))
			else:
				if cls.__name__ not in editor.modes.keys():
					continue
				editor.modes.remove(cls.__name__)
			editor.viewport().repaint()

	def _toggle_panel(self, config, enabled, cls, pos, **kwargs):

		cfg[config] = enabled
		for editor in self._editors:
			if enabled:
				if cls.__name__ in editor.panels._panels[pos]:
					continue
				panel = cls(**kwargs)
				editor.panels.append(panel, position=pos)
				editor.panels.refresh()
				panel.show()
			else:
				if cls.__name__ not in editor.panels._panels[pos]:
					continue
				editor.panels.remove(cls.__name__)
				editor.panels.refresh()



#-*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_element import base_element

class base_rect_ellipse(base_element):

	pos_start = None
	preview = None

	@classmethod
	def mouse_release(cls, sketchpad, pos):

		if cls.pos_start is None:
			return
		w = pos[0]-cls.pos_start[0]
		h = pos[1]-cls.pos_start[1]
		# Don't draw line-like shapes
		if abs(w) < 1 or abs(h) < 1:
			cls.pos_start = None
			sketchpad.canvas.removeItem(cls.preview)
			return
		properties = {
				u'x':		cls.pos_start[0],
				u'y':		cls.pos_start[1],
				u'w':		w,
				u'h':		h,
				u'color': 	sketchpad.current_color(),
				u'penwidth'	: sketchpad.current_penwidth(),
				u'fill'		: sketchpad.current_fill(),
				u'show_if'	: sketchpad.current_show_if()
			}
		e = cls(sketchpad, properties=properties)
		cls.pos_start = None
		sketchpad.canvas.removeItem(cls.preview)
		return e

	@classmethod
	def mouse_move(cls, sketchpad, pos):

		if cls.pos_start is None or cls.preview is None:
			return
		x = cls.pos_start[0]
		y = cls.pos_start[1]
		w = pos[0]-cls.pos_start[0]
		h = pos[1]-cls.pos_start[1]
		if w < 0:
			x += w
			w *= -1
		if h < 0:
			y += h
			h *= -1
		cls.preview.setRect(x, y, w, h)

	@classmethod
	def reset(cls):

		cls.pos_start = None

	@staticmethod
	def requires_color():
		return True

	@staticmethod
	def requires_penwidth():
		return True

	@staticmethod
	def requires_fill():
		return True



# -*- coding: utf-8 -*-
"""
@author: Daniel Schreij

This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

This module is distributed under the Apache v2.0 License.
You should have received a copy of the Apache v2.0 License
along with this module. If not, see <http://www.apache.org/licenses/>.
"""

import QNotifications
import time
import platform
from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from qtpy.QtCore import PYQT_VERSION_STR


__author__ = u"Daniel Schreij"
__license__ = u"GPLv3"


class notifications(base_extension):

	def event_startup(self):

		self.old_notifications = {}
		self.expiration_time = 60  # a minute
		self._suspend = False
		self.notification_area = QNotifications.QNotificationArea(
			self.tabwidget,
			useGlobalCSS=True
		)
		self.notification_area.move(0, 15)
		self.notification_area.setEntryEffect(u'fadeIn', 200)
		self.notification_area.setExitEffect(u'fadeOut', 200)
		# Filthly hack to make this work with Qt4 on the Mac
		# Otherwise the notifications are not shown on top.
		if platform.system() == "Darwin" and PYQT_VERSION_STR < '5':
			self.notification_area.setParent(self.main_window)
			self.notification_area.setParent(self.tabwidget)

	def event_notify(
		self,
		message,
		category='primary',
		timeout=5000,
		always_show=False,
		buttontext=None
	):
		""" Show a notification 'message' in the style 'notification type' for
		'timeout' milliseconds (where 0 milliseconds displays the notification
		indefinitely, until the user removes it)."""
		if self._suspend:
			return
		current_time = time.time()
		# See if notification has been shown before. If it is within
		# self.expiration_time, don't show it again.
		if not always_show:
			if (message, category, timeout, buttontext) in self.old_notifications:
				prev_time = self.old_notifications[(
					message,
					category,
					timeout,
					buttontext
				)]
				if current_time - prev_time < self.expiration_time:
					return
			# Add notification to old notifications list.
			self.old_notifications[(
				message,
				category,
				timeout,
				buttontext
			)] = current_time
		try:
			self.notification_area.display(
				message, category, timeout, timeout != 0, buttontext
			)
		except TypeError:
			# Older versions of QNotifications did not support the autohide
			# option
			self.notification_area.display(
				message, category, timeout, buttontext
			)

	def event_notify_suspend(self):

		self._suspend = True

	def event_notify_resume(self):

		self._suspend = False



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import libopensesame.python_workspace_api as api
from libopensesame.exceptions import AbortCoroutines
import types
from openexp.canvas_elements import ElementFactory
from libopensesame.widgets.widget_factory import WidgetFactory
from libopensesame.base_python_workspace import base_python_workspace


class python_workspace(base_python_workspace):

	"""
	desc:
		Provides a worspace in which Python scripts can be executed. Generally,
		each experiment has one python_workspace object, which is used for all
		inline_script items, and other items that make allow users to execute
		Python code.

		The python_workspace object is dictionary-like, and allows you to get
		and set global variables in a dictionary way, e.g.:

			self.experiment.python_workspace[u'my_var'] = 10
	"""

	def __init__(self, experiment):

		"""
		desc:
			Constructor.

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment
		"""

		base_python_workspace.__init__(self, experiment)
		api.experiment = experiment

	def init_globals(self):

		"""
		desc:
			Initialize the global workspace.
		"""

		# By setting the __name__ global, the workspace will operate as a
		# module, so that e.g. import statements don't suffer from locality.
		self._globals.update({
			u'__name__'			: u'python_workspace',
			u'exp'				: self.experiment,
			u'var'				: self.experiment.var,
			u'pool'				: self.experiment.pool,
			u'items'			: self.experiment.items,
			u'clock'			: self.experiment._clock,
			u'log'				: self.experiment._log,
			u'responses'		: self.experiment._responses,
			u'data_files'		: self.experiment.data_files,
			u'AbortCoroutines'	: AbortCoroutines,
		})
		# All functions from the api modules are also loaded into the globals.
		# This way they can be called directly by name.
		api.set_aliases()
		for name, obj in api.__dict__.items():
			if self._is_api_object(obj):
				self._globals[name] = obj

	def _is_api_object(self, obj):

		if isinstance(obj, types.FunctionType):
			return True
		if isinstance(obj, type) and issubclass(
			obj, (WidgetFactory, ElementFactory)):
				return True
		return False



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import os.path
import platform
import sys
try:
	import site
except IOError:
	# On Android, the site package cannot be imported
	site = None
from libopensesame import metadata
from libopensesame.oslogging import oslogger


def parse_environment_file():

	"""
	desc:
		Parses the environment.yaml file if it exists.
	"""

	if not os.path.exists(u'environment.yaml'):
		return
	with safe_open(u'environment.yaml') as fd:
		d = safe_yaml_load(fd.read())
	# Convert all values from UTF8 to the filesystem encoding
	for key, val in d.items():
		d[key] = safe_str(safe_decode(val), enc=filesystem_encoding())
	# The Python path is added to sys.path, the rest is added as an environment
	# variable.
	if u'PYTHON_PATH' in d:
		sys.path = d[u'PYTHON_PATH'].split(';') + sys.path
		del d[u'PYTHON_PATH']
	os.environ.update(d)

def change_working_dir():

	"""A horrifyingly ugly hack to change the working directory under Windows"""

	import libqtopensesame.qtopensesame

	if os.name == "nt":
		try:
			# Extract the part of the description containing the path
			s = str(libqtopensesame.qtopensesame)
			i = s.find(u'from \'') + 6
			j = s.find(u'\'>\'') - 1
			s = s[i:j]
			# Go up the tree until the path of the current script
			while not os.path.exists(os.path.join(s, u'opensesame')) and \
				not os.path.exists(os.path.join(s, u'opensesame.exe'))and \
				not os.path.exists(os.path.join(s, u'opensesamerun')) and \
				not os.path.exists(os.path.join(s, u'opensesamerun.exe')):
				s = os.path.dirname(s)
			os.chdir(s)
			if s not in sys.path:
				sys.path.append(s)
		except Exception as e:
			oslogger.debug(u'failed to change working directory: %s' % e)

def opensesamerun_options():

	"""Parse the command line options for opensesamerun"""

	import optparse

	parser = optparse.OptionParser(
		u'usage: opensesamerun [experiment] [options]', version=u'%s \'%s\'' % \
		(metadata.__version__, metadata.codename))
	parser.set_defaults(subject=0)
	parser.set_defaults(logfile=None)
	parser.set_defaults(debug=False)
	parser.set_defaults(fullscreen=False)
	parser.set_defaults(pylink=False)
	parser.set_defaults(width=1024)
	parser.set_defaults(height=768)
	parser.set_defaults(custom_resolution=False)
	group = optparse.OptionGroup(parser, u'Subject and log file options')
	group.add_option(u"-s", u"--subject", action=u"store", dest=u"subject", \
		help=u"Subject number")
	group.add_option(u"-l", u"--logfile", action=u"store", dest=u"logfile", \
		help=u"Logfile")
	parser.add_option_group(group)
	group = optparse.OptionGroup(parser, u"Display options")
	group.add_option(u"-f", u"--fullscreen", action=u"store_true", dest= \
		"fullscreen", help=u"Run fullscreen")
	group.add_option(u"-c", u"--custom_resolution", action=u"store_true", \
		dest=u"custom_resolution", help= \
		u"Do not use the display resolution specified in the experiment file")
	group.add_option(u"-w", u"--width", action=u"store", dest=u"width", help= \
		u"Display width")
	group.add_option(u"-e", u"--height", action=u"store", dest=u"height", \
		help=u"Display height")
	parser.add_option_group(group)
	group = optparse.OptionGroup(parser, u"Miscellaneous options")
	group.add_option(u"-d", u"--debug", action=u"store_true", dest=u"debug", \
		help=u"Print lots of debugging messages to the standard output")
	group.add_option(u"--stack", action=u"store_true", dest=u"stack", help= \
		u"Print stack information")
	parser.add_option_group(group)
	group = optparse.OptionGroup(parser, u"Miscellaneous options")
	group.add_option(u"--pylink", action=u"store_true", dest=u"pylink", help= \
		u"Load PyLink before PyGame (necessary for using the Eyelink plug-ins in non-dummy mode)")
	parser.add_option_group(group)
	options, args = parser.parse_args(sys.argv)

	# Set the default logfile based on the subject nr
	if options.logfile is None:
		options.logfile = u"subject-%s.csv" % options.subject

	if len(sys.argv) > 1 and os.path.exists:
		options.experiment = sys.argv[1]
	else:
		options.experiment = u""

	try:
		options.subject = int(options.subject)
	except:
		parser.error(u"Subject (-s / --subject) should be numeric")

	try:
		options.width = int(options.width)
	except:
		parser.error(u"Width (-w / --width) should be numeric")

	try:
		options.height = int(options.height)
	except:
		parser.error(u"Height (-e / --height) should be numeric")

	return options

def opensesamerun_ready(options):

	"""
	Check if the opensesamerun options are sufficiently complete to run the
	experiment

	Arguments:
	options -- a dictionary containing the options

	Returns:
	True or False, depending on whether the options are sufficient
	"""

	# Check if the experiment exists
	if not os.path.exists(options.experiment):
		return False

	# Check if the logfile is writable
	try:
		open(options.logfile, u"w")
	except:
		return False

	# Ready to run!
	return True

def messagebox(title, msg):

	"""
	Presents a simple tk messagebox

	Arguments:
	title -- the title of the messagebox
	msg -- the message
	"""

	if py3:
		import tkinter as Tkinter
	else:
		import Tkinter
	root = Tkinter.Tk()
	root.title(title)
	l = Tkinter.Label(root, text=msg, justify=Tkinter.LEFT, padx=8, pady=8,
		wraplength=300)
	l.pack()
	b = Tkinter.Button(root, text=u"Ok", command=root.quit)
	b.pack(side=Tkinter.RIGHT)
	root.mainloop()

def strip_tags(s):

	"""
	Strip html tags from a string and convert breaks to newlines.

	Arguments:
	s -- the string to be stripped

	Returns:
	The stripped string
	"""

	import re
	return re.compile(r'<.*?>').sub('', str(s).replace("<br />", \
		"\n").replace("<br>", "\n"))

def resource(name):

	"""
	A hacky way to get a resource using the functionality from openexp

	Arguments:
	name	--	The name of the requested resource. If this is a regular string
				it is assumed to be encoded as utf-8.

	Returns:
	A Unicode string with the full path to the resource.
	"""

	for folder in base_folders:
		path = os.path.join(folder, u'opensesame_resources',
			safe_decode(name, enc=u'utf-8'))
		if os.path.exists(path):
			return path
	return None

def home_folder():

	"""
	Determines the home folder.

	Returns:
	A path to the home folder.
	"""

	import platform
	if platform.system() == u"Windows":
		home_folder = os.environ[u"APPDATA"]
	elif platform.system() == u"Darwin":
		home_folder = os.environ[u"HOME"]
	elif platform.system() == u"Linux":
		home_folder = os.environ[u"HOME"]
	else:
		home_folder = os.environ[u"HOME"]
	return safe_decode(home_folder, enc=filesystem_encoding())

def opensesame_folder():

	"""
	Determines the folder that contains the OpenSesame executable. This is only
	applicable under Windows.

	Returns:
	The OpenSesame folder or None if the os is not Windows.
	"""
	# Determines the directory name of the script or the directory name
	# of the executable after being packaged with py2exe. This has to be
	# done so the child process can find all relevant modules too.
	# See http://www.py2exe.org/index.cgi/HowToDetermineIfRunningFromExe
	#
	# There are three scenarios:
	#
	# - OpenSesame is run from a frozen state, in which case the OpenSesame
	#   folder is the folder containing the
	# - OpenSesame is run from source, in which case we go to the OpenSesame
	#   folder by going a levels up from the __file__ folder.
	# - OpenSesame is run in Anaconda, in which case we need to go two levels
	#   up to get out of the site-packages folder.
	if platform.system() == u'Darwin':
		return os.getcwd()
	elif platform.system() == u'Windows':
		import imp
		if (
			hasattr(sys, u'frozen') or
			hasattr(sys, u'importers') or
			imp.is_frozen(u'__main__')
		):
			path = safe_decode(
				os.path.dirname(sys.executable),
				enc=sys.getfilesystemencoding()
			)
		else:
			# To get the opensesame folder, simply jump to levels up
			path = safe_decode(
				os.path.dirname(__file__),
				enc=sys.getfilesystemencoding()
			)
			# The current should not be the site-packages folder, which
			# happens on Anaconda if launched in multiprocess mode.
			path = os.path.normpath(os.path.join(path, u'..'))
			if path.endswith(u'Lib\\site-packages'):
				path = os.path.normpath(os.path.join(path, u'..', u'..'))
		return path
	else:
		return None

def open_url(url):

	"""
	Open a URL in an OS specific way. The URL can be a file, website, etc.

	Arguments:
	url -- a url
	"""

	oslogger.info(u'opening %s' % url)
	if url.startswith(u'http://') or url.startswith(u'https://'):
		import webbrowser
		webbrowser.open(url)
		return
	import platform
	import subprocess
	if platform.system() == u"Linux":
		pid = subprocess.Popen([u"xdg-open", url]).pid
	elif platform.system() == u"Darwin":
		pid = subprocess.Popen(["open", url]).pid
	elif platform.system() == u"Windows":
		try:
			os.startfile(url)
		except:
			oslogger.error(u"Failed to open '%s'" % url)
	else:
		oslogger.error(u"Failed to open '%s'" % url)

def filesystem_encoding():

	"""
	Gets the current file system encoding. This wrapper is necessary, because
	sys.getfilesystemencoding() returns None on Android.

	Returns:
	A string with the file system encoding, such as 'utf-8' or 'mdcs'
	"""

	enc = sys.getfilesystemencoding()
	if enc is None:
		enc = u'utf-8'
	return enc

def strip_html(s):

	"""
	Strips basic HTML tags from a string.

	Arguments:
	s		--	A string to strip.

	Returns:
	A stripped string.
	"""

	s = s.replace(u'<br />', u'\n')
	for tag in [u'<i>', u'</i>', u'<b>', u'</b>']:
		s = s.replace(tag, u'')
	return s

def escape_html(s):

	"""
	Escapes a string so that it can be displayed as HTML. This is useful for
	example for tracebacks, which use <> characters.

	Arguments:
	s	--	A string to escape. We assume Unicode input. str objects may cause
			decoding errors.

	Returns:
	An escaped string.
	"""

	# Note that we need to replace the '&' first, otherwise we'll start escaping
	# the escaped characters.
	l = [(u'&', u'&amp;'), (u' ', u'&nbsp;'), (u'\t', \
		u'&nbsp;&nbsp;&nbsp;&nbsp;'), (u'<', u'&lt;'), (u'>', u'&gt;')]
	for orig, new in l:
		s = s.replace(orig, new)
	return s


# Build a list of base folders, that is, folders that can have
# opensesame_resources, opensesame_plugins, or opensesame_extensions as
# subfolders. Which folders are available depends on the environment, but the
# following folders are scanned:
#
# - The working directory (cwd)
# - The folder that contains libopensesame (parent_folder)
# - The share folder within an Anaconda/ Miniconda environment
# - The userbase folder:
#		- ~/.local (on Ubuntu 16.04)
# - The user site-packages folder:
#		- ~/.local/lib/python[X]/site-packages (on Ubuntu 16.04)
# - The global site-packages folders (on Ubuntu 16.04):
#		- /usr/local/lib/python3.5/dist-packages
#		- /usr/lib/python3/dist-packages
#		- /usr/lib/python3.5/dist-packages
# - /usr/local/share
# - /usr/share

if py3:
	cwd = os.getcwd()
else:
	cwd = os.getcwdu()
parent_folder = safe_decode(
	os.path.dirname(os.path.dirname(__file__)),
	enc=filesystem_encoding()
)
base_folders = [cwd, parent_folder]
# Locate Anaconda/Miniconda share
base_folders.append(os.path.join(
	safe_decode(
		os.path.dirname(os.path.dirname(sys.executable)),
		enc=filesystem_encoding()
	),
	u"share"
))
if hasattr(site, u'getuserbase'):
	base_folders.append(os.path.join(
		safe_decode(site.getuserbase(), enc=filesystem_encoding()),
		u'share'
	))
if hasattr(site, u'getusersitepackages'):
	base_folders.append(os.path.join(
		safe_decode(site.getusersitepackages(), enc=filesystem_encoding()),
		u'share'
	))
if hasattr(site, u'getsitepackages'):
	base_folders += [
		os.path.join(
			safe_decode(folder, enc=filesystem_encoding()),
			u'share'
		)
		for folder in site.getsitepackages()
	]
base_folders += [u'/usr/local/share', u'/usr/share']
base_folders = list(filter(os.path.exists, base_folders))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension, suspend_events
from libqtopensesame.items.qtstructure_item import qtstructure_item
from qtpy.QtWidgets import QMenu, QToolBar
from undo_stack import undo_stack
import difflib
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'undo_manager', category=u'extension')

class undo_manager(base_extension):

	"""
	desc:
		An extension that implements undo/ redo.
	"""

	def event_startup(self):

		"""
		desc:
			Initializes the extension on startup.
		"""

		self.undo_action = self.qaction(u'edit-undo', _(u'Undo'), self.undo,
			tooltip=_(u'Undo most recent action'),
			shortcut=u'Alt+Ctrl+Z')
		for w in self.action.associatedWidgets():
			if not isinstance(w, QMenu) and not isinstance(w, QToolBar):
				continue
			w.insertAction(self.action, self.undo_action)
		self.initialize()

	def event_open_experiment(self, path):

		"""
		desc:
			Re-initializes when a new experiment is opened.
		"""

		self.initialize()

	def event_change_item(self, name):

		"""
		desc:
			Remember an items state when it has changed.
		"""

		self.remember_item_state(name)

	def event_new_item(self, name, _type):

		"""
		desc:
			Remember addition of a new item.
		"""

		self.remember_new_item(name)

	def activate(self):

		"""
		desc:
			Activated the redo action (undo is added as custom action in
			event_startup).
		"""

		self.redo()

	def initialize(self):

		"""
		desc:
			Initializes an empty undo stack.
		"""

		self.stack = undo_stack()
		for item in self.experiment.items:
			self.remember_item_state(item)
		self.undo_action.setEnabled(self.stack.can_undo())

	def remember_experiment_state_before(self):

		"""
		desc:
			Remember the entire experiment state before it has changed.
		"""

		self.remember_experiment_state()

	def remember_experiment_state_after(self):

		"""
		desc:
			Remember the entire experiment state after it has changed, and
			update the current-item status. If no change has actually occurred
			the last two experiments are discarded.
		"""

		self.remember_experiment_state()
		if self.prune_unchanged_experiment():
			return
		for item in self.experiment.items:
			self.make_current(item)

	def make_current(self, item):

		"""
		desc:
			Update the current-item status for one item.

		arguments:
			item:	The item name.
		"""

		script = self.experiment.items[item].cached_script
		self.stack.set_current(item, script)

	def remember_item_state(self, item):

		"""
		desc:
			Push the state of an item to the undo stack.

		arguments:
			item:	The item name.
		"""

		script = self.experiment.items[item].cached_script
		self.stack.add(item, script)
		self.set_enabled(self.stack.can_redo())
		self.undo_action.setEnabled(self.stack.can_undo())

	def remember_experiment_state(self):

		"""
		desc:
			Push the state of the entire experiment to the undo stack.
		"""

		script = self.experiment.to_string()
		self.stack.add(u'__experiment__', script)
		self.set_enabled(self.stack.can_redo())
		self.undo_action.setEnabled(self.stack.can_undo())

	def remember_new_item(self, name):

		self.stack.add(u'__newitem__', name)
		self.set_enabled(self.stack.can_redo())
		self.undo_action.setEnabled(self.stack.can_undo())

	def prune_unchanged_experiment(self):

		"""
		desc:
			If the last two items on the undo stack are full experiments, and
			there is no change between them, discard them.
		"""

		if len(self.stack) < 2:
			return False
		key1, state1 = self.stack.peek(-1)
		key2, state2 = self.stack.peek(-2)
		if key1 == key2 == u'__experiment__' and state1 == state2:
			self.stack.history = self.stack.history[:-2]
			return True
		return False

	@suspend_events
	def undo(self, dummy=False):

		"""
		desc:
			Perform an undo action.
		"""

		if not self.stack:
			return
		item, script = self.stack.undo()
		if item == u'__experiment__':
			self.main_window.regenerate(script)
		elif item == u'__newitem__':
			del self.experiment.items[script]
			self.tabwidget.close_all()
		else:
			if item not in self.experiment.items:
				return
			self.experiment.items[item].from_string(script)
			self.experiment.items[item].update()
			self.experiment.items[item].open_tab()
			if isinstance(self.experiment.items[item], qtstructure_item):
				self.experiment.items.clear_cache()
				self.experiment.build_item_tree()
		self.set_enabled(self.stack.can_redo())
		self.undo_action.setEnabled(self.stack.can_undo())

	@suspend_events
	def redo(self):

		"""
		desc:
			Perform a redo action.
		"""

		item, script = self.stack.redo()
		if item not in self.experiment.items:
			return
		self.experiment.items[item].from_string(script)
		self.experiment.items[item].update()
		self.experiment.items[item].open_tab()
		if isinstance(self.experiment.items[item], qtstructure_item):
			self.experiment.items.clear_cache()
			self.experiment.build_item_tree()
		self.set_enabled(self.stack.can_redo())
		self.undo_action.setEnabled(self.stack.can_undo())

	def show(self):

		"""
		desc:
			Print a summary of undo actions to the debug window.
		"""

		item, old_script = self.stack.peek()
		if item is None:
			print(u'\nThe undo stack is empty')
			return
		self.console.write(u'\nMost recent: %s\n\n' % item)
		if item == u'__experiment__':
			old_script = self.stack.peek(-2)[1]
			new_script = self.stack.peek(-1)[1]
		elif item == u'__newitem__':
			new_script = old_script
		else:
			new_script = self.stack.current[item][0]
		for line in difflib.ndiff(old_script.splitlines(),
			new_script.splitlines()):
			if line.startswith(u'+'):
				self.console.write(u'\x1b[32;1m')
			elif line.startswith(u'-'):
				self.console.write(u'\x1b[31;1m')
			else:
				continue
			self.console.write(line + u'\n')
		self.console.write(u'\x1b[0m')
		self.console.write(u'\nFull stack:\n\n')
		for i, (item, script) in enumerate(self.stack.history[::-1]):
			self.console.write(u'%d - %s\n' % (i, item))

	# All these events simply undo by restoring the complete experiment state.
	# This is crude, but works for now.

	def event_prepare_regenerate(self):
		self.extension_manager.suspend_until(u'regenerate')
		self.remember_experiment_state_before()

	def event_regenerate(self):
		self.remember_experiment_state_after()

	def event_prepare_delete_item(self, name):
		self.extension_manager.suspend_until(u'delete_item')
		self.remember_experiment_state_before()

	def event_delete_item(self, name):
		self.remember_experiment_state_after()

	def event_prepare_rename_item(self, from_name, to_name):
		self.extension_manager.suspend_until(u'rename_item')
		self.remember_experiment_state_before()

	def event_rename_item(self, from_name, to_name):
		self.remember_experiment_state_after()

	def event_prepare_purge_unused_items(self):
		self.extension_manager.suspend_until(u'purge_unused_items')
		self.remember_experiment_state_before()

	def event_purge_unused_items(self):
		self.remember_experiment_state_after()

	def event_prepare_change_experiment(self):
		self.extension_manager.suspend_until(u'change_experiment')
		self.remember_experiment_state_before()

	def event_change_experiment(self):
		self.remember_experiment_state_after()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import sys
from libopensesame import misc, metadata
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc import template_info
_ = translation_context(u'get_started', category=u'extension')

class get_started(base_extension):

	"""
	desc:
		Shows the get-started tab and opens an experiment on startup, if one was
		passed on the command line.
	"""

	def activate(self):

		"""
		desc:
			Is called when the extension is activated through the menu/ toolbar
			action.
		"""

		# Initialize templates
		templates = []
		for i, (path, desc) in enumerate(template_info.templates):
			try:
				path = self.experiment.resource(path)
			except:
				continue
			if not i:
				cls = u'important-button'
			else:
				cls = u'button'
			# We need to use forward slashes, also on Windows, otherwise it
			# leads to an invalid URL when prefixed with opensesame://
			path = os.path.abspath(path).replace(u'\\', u'/')
			md = u'<a href="opensesame://%s" class="%s">%s</a><br />' \
				% (path, cls, desc)
			templates.append(md)
		# Initialize recent experiments
		if not self.main_window.recent_files:
			recent = []
		else:
			recent = [_(u'Continue with a recent experiment:')+u'<br />']
			for i, path in enumerate(self.main_window.recent_files):
				cls = u'important-button' if not i else u'button'
				md = u'<a href="opensesame://event.open_recent_%d" class="%s">%s</a><br />' % \
					(i, cls, self._unambiguous_path(path))
				recent.append(md)
		# Create markdown
		with safe_open(self.ext_resource(u'get_started.md')) as fd:
			md = fd.read()
		md = md % {
			u'version' : metadata.__version__,
			u'codename' : metadata.codename,
			u'templates' : u'  \n'.join(templates),
			u'recent_experiments' : u'  \n'.join(recent)
			}
		self.tabwidget.open_markdown(md, title=_(u'Get started!'),
			icon=u'document-new')

	def _unambiguous_path(self, path):

		"""
		desc:
			If the path basename is unique among the resent experiments, this is
			used. Otherwise, the full path is used.

		arguments:
			path:	The path to shorten unambiguously.

		returns:
			The unambiguously shortened path.
		"""

		basename = os.path.basename(path)
		basenames = \
			[os.path.basename(_path) for _path in self.main_window.recent_files]
		return path if basenames.count(basename) > 1 else basename

	def event_open_recent_0(self):
		self.main_window.open_file(path=self.main_window.recent_files[0])
	def event_open_recent_1(self):
		self.main_window.open_file(path=self.main_window.recent_files[1])
	def event_open_recent_2(self):
		self.main_window.open_file(path=self.main_window.recent_files[2])
	def event_open_recent_3(self):
		self.main_window.open_file(path=self.main_window.recent_files[3])
	def event_open_recent_4(self):
		self.main_window.open_file(path=self.main_window.recent_files[4])

	@base_extension.as_thread(wait=500)
	def event_startup(self):

		"""
		desc:
			Called on startup.
		"""

		# Open an experiment if it has been specified as a command line argument
		# and suppress the new wizard in that case.
		if len(sys.argv) >= 2 and os.path.isfile(sys.argv[1]):
			path = safe_decode(sys.argv[1], enc=misc.filesystem_encoding(),
				errors=u'ignore')
			self.main_window.open_file(path=path)
			return
		self.activate()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'item_view_button', category=u'core')

class item_view_button(base_subcomponent, QtWidgets.QPushButton):

	"""
	desc:
		The view button in the top-right of the edit controls.
	"""

	def __init__(self, item):

		"""
		desc:
			Constructor.

		arguments:
			item:
				desc:	An item.
				type:	qtitem.
		"""

		super(item_view_button, self).__init__(item.main_window)
		self.item = item
		self.setup(item)
		self.set_view_icon(u'controls')
		self.setIconSize(QtCore.QSize(16,16))
		self.menu_view = QtWidgets.QMenu()
		self.menu_view.addAction(self.view_controls_icon(),
			_(u'View controls'), self.item.set_view_controls)
		self.menu_view.addAction(self.view_script_icon(), _(u'View script'),
			self.item.set_view_script)
		self.menu_view.addAction(self.view_split_icon(), _(u'Split view'),
			self.item.set_view_split)
		self.setMenu(self.menu_view)
		self.setToolTip(_(u'Select view'))

	def view_controls_icon(self):

		"""
		returns:
			desc:	The icon for the controls view.
			type: 	QIcon
		"""

		return self.theme.qicon(u'os-view-controls')

	def view_script_icon(self):

		"""
		returns:
			desc:	The icon for the script view.
			type: 	QIcon
		"""

		return self.theme.qicon(u'os-view-script')

	def view_split_icon(self):

		"""
		returns:
			desc:	The icon for the split view.
			type: 	QIcon
		"""

		return self.theme.qicon(u'os-view-split')

	def set_view_icon(self, view):

		"""
		desc:
			Sets the icon of the button according to a view.

		arguments:
			view:
				desc:	'controls', 'script', or 'split'.
				type:	unicode
		"""

		if view == u'controls':
			self.setIcon(self.view_controls_icon())
		elif view == u'script':
			self.setIcon(self.view_script_icon())
		else:
			self.setIcon(self.view_split_icon())



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from qtpy import QtWidgets
from libqtopensesame.widgets.general_header_widget import general_header_widget
from libqtopensesame.widgets.base_widget import base_widget
from openexp._color.color import color
from openexp import backend
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'general_properties', category=u'core')


class general_properties(base_widget):

	"""The QWidget for the general properties tab."""

	backend_format = u'%s [%s]'

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	A qtopensesame object.
		"""

		super(general_properties, self).__init__(
			main_window,
			ui=u'widgets.general_properties'
		)
		self.lock = False
		# Set the header, with the icon, label and script button
		self.header_widget = general_header_widget(self, self.main_window)
		header_hbox = QtWidgets.QHBoxLayout()
		header_hbox.addWidget(self.theme.qlabel(u"experiment"))
		header_hbox.addWidget(self.header_widget)
		header_hbox.addStretch()
		header_hbox.setContentsMargins(0, 0 ,0, 0)
		header_hbox.setSpacing(12)
		header_widget = QtWidgets.QWidget()
		header_widget.setLayout(header_hbox)
		self.ui.container_layout.insertWidget(0, header_widget)
		# Initialize the color and font widgets
		self.ui.edit_foreground.initialize(self.experiment)
		self.ui.edit_background.initialize(self.experiment)
		self.ui.widget_font.initialize(self.experiment)
		# Set the backend combobox
		for name, info in backend.backend_info(self.experiment).items():
			desc = info[u"description"]
			icon = info[u"icon"]
			self.ui.combobox_backend.addItem(self.main_window.theme.qicon(
				icon), self.backend_format % (name, desc))
		self.quick_connect(
			slot=self.main_window.ui.tabwidget.open_general_script,
			signals=[self.ui.button_script_editor.clicked],
		)
		self.quick_connect(
			slot=self.main_window.ui.tabwidget.open_backend_settings,
			signals=[self.ui.button_backend_settings.clicked]
		)
		self.quick_connect(
			slot=self.apply_changes,
			signals=[
				self.ui.combobox_backend.currentIndexChanged,
				self.ui.spinbox_width.editingFinished,
				self.ui.spinbox_height.editingFinished,
				self.ui.checkbox_disable_garbage_collection.stateChanged,
				self.ui.edit_foreground.textEdited,
				self.ui.edit_background.textEdited,
				self.ui.widget_font.font_changed,
			])
		self.tab_name = u'__general_properties__'
		self.on_activate = self.refresh

	def set_header_label(self):

		"""
		desc:
			Sets the general header based on the experiment title and
			description.
		"""

		self.header_widget.set_name(self.experiment.var.title)
		self.header_widget.set_desc(self.experiment.var.description)

	def apply_changes(self):

		"""
		desc:
			Applies changes to the general tab.
		"""

		# Skip if the general tab is locked and lock it otherwise
		if self.lock:
			return
		self.lock = True

		self.main_window.set_busy(True)
		self.main_window.extension_manager.fire(u'prepare_change_experiment')
		# Set the title and the description
		title = self.experiment.syntax.sanitize(
			self.header_widget.edit_name.text())
		if title != self.experiment.var.title:
			self.experiment.var.title = title
			self.experiment.build_item_tree()
		desc = self.experiment.syntax.sanitize(
			self.header_widget.edit_desc.text())
		self.experiment.var.description = desc
		# Set the backend
		if self.ui.combobox_backend.isEnabled():
			i = self.ui.combobox_backend.currentIndex()
			_backend = list(backend.backend_info(self.experiment).values())[i]
			self.experiment.var.canvas_backend = _backend[u"canvas"]
			self.experiment.var.keyboard_backend = _backend[u"keyboard"]
			self.experiment.var.mouse_backend = _backend[u"mouse"]
			self.experiment.var.sampler_backend = _backend[u"sampler"]
			self.experiment.var.clock_backend = _backend[u"clock"]
			self.experiment.var.color_backend = _backend[u"color"]
		else:
			oslogger.debug(
				u'not setting back-end, because a custom backend is selected'
			)
		# Set the display width
		width = self.ui.spinbox_width.value()
		height = self.ui.spinbox_height.value()
		if self.experiment.var.width != width or \
			self.experiment.var.height != height:
			self.main_window.update_resolution(width, height)
		# Set the foreground color
		foreground = self.experiment.syntax.sanitize(self.ui.edit_foreground.text())
		refs = []
		try:
			refs = self.experiment.get_refs(foreground)
			color.to_hex(foreground)
		except Exception as e:
			if refs == []:
				self.experiment.notify(e)
				foreground = self.experiment.var.foreground
				self.ui.edit_foreground.setText(foreground)
		self.experiment.var.foreground = foreground
		# Set the background color
		background = self.experiment.syntax.sanitize(self.ui.edit_background.text())
		refs = []
		try:
			refs = self.experiment.get_refs(background)
			color.to_hex(background)
		except Exception as e:
			if refs == []:
				self.experiment.notify(e)
				background = self.experiment.var.background
				self.ui.edit_background.setText(background)
		self.experiment.var.background = foreground
		self.experiment.var.background = background
		# Set the font
		self.experiment.var.font_family = self.ui.widget_font.family
		self.experiment.var.font_size = self.ui.widget_font.size
		self.experiment.var.font_italic = self.ui.widget_font.italic
		self.experiment.var.font_bold = self.ui.widget_font.bold
		# Other checkboxes
		self.experiment.var.disable_garbage_collection = \
			self.ui.checkbox_disable_garbage_collection.isChecked()
		# Refresh the interface and unlock the general tab
		self.lock = False
		self.main_window.extension_manager.fire(u'change_experiment')
		self.main_window.set_busy(False)

	def refresh(self):

		"""
		desc:
			Updates the controls of the general tab.
		"""

		# Lock the general tab to prevent a recursive loop
		self.lock = True
		# Set the header containing the titel etc
		self.set_header_label()
		# Select the backend
		_backend = backend.backend_match(self.experiment)
		if _backend == u"custom":
			self.ui.combobox_backend.setDisabled(True)
		else:
			self.ui.combobox_backend.setDisabled(False)
			desc = backend.backend_info(self.experiment)[_backend]\
				[u"description"]
			i = self.ui.combobox_backend.findText(
				self.backend_format % (_backend, desc))
			self.ui.combobox_backend.setCurrentIndex(i)
		# Set the resolution
		try:
			self.ui.spinbox_width.setValue(int(self.experiment.var.width))
			self.ui.spinbox_height.setValue(int(self.experiment.var.height))
		except ValueError:
			self.experiment.notify(
				_(u"Failed to parse the resolution. Expecting positive numeric values.")
			)
		# Set the colors
		self.ui.edit_foreground.setText(safe_decode(
			self.experiment.var.foreground))
		self.ui.edit_background.setText(safe_decode(
			self.experiment.var.background))
		self.ui.widget_font.initialize(self.experiment)
		self.ui.checkbox_disable_garbage_collection.setChecked(
			self.experiment.var.disable_garbage_collection == u'yes')
		# Release the general tab
		self.lock = False



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.config import cfg
from variable_inspector_dockwidget import variable_inspector_dockwidget
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'variable_inspector', category=u'extension')

class variable_inspector(base_extension):

	"""
	desc:
		A variable inspector.
	"""

	def event_startup(self):

		"""
		desc:
			Initializes the variable inspector dock widget.
		"""

		self.need_refresh = False
		self.dock_widget = variable_inspector_dockwidget(self.main_window, self)
		self.dock_widget.visibilityChanged.connect(self.set_visible)
		self.main_window.addDockWidget(
			QtCore.Qt.RightDockWidgetArea,
			self.dock_widget
		)
		self.set_visible(cfg.variable_inspector_visible)
		self.shortcut_focus = QtWidgets.QShortcut(QtGui.QKeySequence(
			cfg.variable_inspector_focus_shortcut), self.main_window,
			self.focus, context=QtCore.Qt.ApplicationShortcut)

	def focus(self):

		"""
		desc:
			Makes the dock visible and sets the focus to the filter box.
		"""

		self.set_visible(True)
		self.dock_widget.widget().focus()

	def open_help(self):

		"""
		desc:
			Opens the help tab.
		"""

		self.tabwidget.open_osdoc('manual/variables')

	def set_visible(self, visible):

		"""
		desc:
			Sets the visibility of the dock widget.

		arguments:
			visible:
				type:	bool
		"""

		cfg.variable_inspector_visible = visible
		self.set_checked(visible)
		if visible:
			if self.need_refresh:
				self.dock_widget.widget().refresh()
				self.need_refresh = False
			self.dock_widget.show()
			self.dock_widget.widget().focus()
		else:
			self.dock_widget.hide()

	def activate(self):

		"""
		desc:
			Toggles the visibility of the dock widget.
		"""

		self.set_visible(not cfg.variable_inspector_visible)

	def refresh(self):

		"""
		desc:
			Refreshes the variable inspector.
		"""

		if self.dock_widget.isVisible():
			self.dock_widget.widget().refresh()
			self.need_refresh = False
		else:
			self.need_refresh = True

	# The following events all refresh the variable inspector

	def event_heartbeat(self):
		self.refresh()

	def event_change_item(self, name):
		self.refresh()

	def event_pause_experiment(self):
		self.refresh()

	def event_set_workspace_globals(self, global_dict):
		self.dock_widget.widget().set_workspace_globals(global_dict)
		self.refresh()

	def event_open_experiment(self, path):
		self.dock_widget.widget().set_workspace_globals({})
		self.refresh()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from libopensesame import plugins
from qtpy import QtWidgets, QtGui
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_append_menu', category=u'core')

class append_existing_action(base_subcomponent, QtWidgets.QAction):

	"""
	desc:
		An action for appending existing items, either as linked or unlinked
		copies.
	"""

	def __init__(self, append_menu, menu, item_name):

		"""
		desc:
			Constructor.

		arguments:
			append_menu:
				desc:	The main menu.
				type:	tree_append_menu
			menu:
				desc:	The submenu for this action.
				type:	QMenu
			item_name:
				desc:	The name of the item to be appended.
				type:	unicode
		"""

		super(append_existing_action, self).__init__(menu)
		self.setup(append_menu)
		self.setText(item_name)
		self.item_name = item_name
		self.setIcon(
			self.theme.qicon(self.experiment.items[item_name].item_icon()))

	def append_item(self, target_item_name):

		"""
		desc:
			Performs the append operation.

		arguments:
			target_item_name:
				desc:	The name of the item (normally a sequence) to which the
						append should be applied.
				type:	unicode
		"""

		target_item = self.experiment.items[target_item_name]
		item_name = self.item_name
		target_item.insert_child_item(item_name, len(target_item.items))

class append_new_action(base_subcomponent, QtWidgets.QAction):

	"""
	desc:
		An action for appending a new item.
	"""

	def __init__(self, append_menu, menu, item_type):

		"""
		desc:
			Constructor.

		arguments:
			append_menu:
				desc:	The main menu.
				type:	tree_append_menu
			menu:
				desc:	The submenu for this action.
				type:	QMenu
			item_type:
				desc:	The type of item to be created and appended.
				type:	unicode
		"""

		super(append_new_action, self).__init__(menu)
		self.setup(append_menu)
		self.item_type = item_type
		self.setText(item_type)
		if item_type not in self.experiment.core_items:
			icon = self.theme.qicon(plugins.plugin_icon_small(item_type))
		else:
			icon = self.theme.qicon(item_type)
		self.setIcon(icon)

	def append_item(self, target_item_name):

		"""
		desc:
			Performs the append operation.

		arguments:
			target_item_name:
				desc:	The name of the item (normally a sequence) to which the
						append should be applied.
				type:	unicode
		"""

		target_item = self.experiment.items[target_item_name]
		item = self.experiment.items.new(self.item_type, catch_exceptions=False)
		if item is not None:
			target_item.insert_child_item(item.name, len(target_item.items))

class tree_append_menu(base_subcomponent, QtWidgets.QMenu):

	"""
	desc:
		An append item menu.
	"""

	def __init__(self, tree_overview, target_treeitem=None):

		"""
		desc:
			Constructor.

		arguments:
			tree_overview:
				desc:	The tree_overview with which this menu is associated.
				type:	tree_overview

		keywords:
			target_treeitem:
				desc:	The treewidget item which corresponds to the sequence to
						which the append operation should be applied. If `None`,
						the top-level item from the tree_overview is used.
				type:	tree_item_item
		"""

		super(tree_append_menu, self).__init__(tree_overview)
		self.setup(tree_overview)
		self.target_treeitem = target_treeitem
		self.tree_overview = tree_overview
		self.action_new_items = QtWidgets.QAction(self.theme.qicon(u'list-add'),
			_(u'Append new item'), self)
		self.addAction(self.action_new_items)
		self.action_linked_copy = QtWidgets.QAction(self.theme.qicon(u'edit-copy'),
			_(u'Append existing item (linked)'), self)
		self.addAction(self.action_linked_copy)
		self.aboutToShow.connect(self.refresh)
		self.triggered.connect(self.append_item)
		self._new_items_menu = None

	def append_item(self, action):

		"""
		desc:
			Performs the append action.

		arguments:
			action:
				desc:	A append_existing_action or append_new_action.
				type:	QAction
		"""

		target_item_name, target_item_ancestry = self.target_treeitem.ancestry()
		action.append_item(target_item_name)
		self.tree_overview.structure_change.emit()
		if self.tree_overview.overview_mode:
			# In overview mode, we need to explicitly update the target item,
			# because this is not linked to the structure_change signal.
			self.experiment.items[target_item_name].update()

	def refresh(self):

		"""
		desc:
			Refreshes the menu before it is shown. This is necessary because the
			structure of the experiment may change.
		"""

		self._items = None
		if self.target_treeitem is None:
			self.target_treeitem = self.tree_overview.topLevelItem(0)
		self.action_linked_copy.setMenu(self.existing_items_menu())
		self.action_new_items.setMenu(self.new_items_menu())

	def existing_items_menu(self):

		"""
		desc:
			Generates a menu with existing items.

		keywords:
			linked:
				desc:	Indicates whether the actions should result in linked
						(True) or unliked (False) copies.
				type:	bool

		returns:
			type:	QMenu
		"""

		self._items = []
		target_item_name = self.target_treeitem.item.name
		for item_name in self.experiment.items:
			if target_item_name in \
				self.experiment.items[item_name].children() or \
				item_name == target_item_name:
				continue
			self._items.append(item_name)
		self._items.sort()
		m = QtWidgets.QMenu(self)
		for item_name in self._items:
			m.addAction(append_existing_action(self, m, item_name))
		return m

	def new_items_menu(self):

		"""
		desc:
			Generates a menu with new items.

		returns:
			type:	QMenu
		"""

		if self._new_items_menu is not None:
			return self._new_items_menu
		self._new_items_menu = QtWidgets.QMenu(self)
		for item_type in self.experiment.core_items + [None] + \
			sorted(plugins.list_plugins()):
			if item_type is None:
				self._new_items_menu.addSeparator()
			else:
				self._new_items_menu.addAction(
					append_new_action(self, self._new_items_menu, item_type))
		return self._new_items_menu



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.var_store import var_store
import warnings
from libopensesame.exceptions import osexception
from libopensesame.oslogging import oslogger


class item(object):

	"""Abstract class that serves as the basis for all OpenSesame items."""

	encoding = u'utf-8'
	var = None
	
	def __init__(self, name, experiment, string=None):

		"""
		Constructor.

		Arguments:
		name 		--	The name of the item.
		experiment 	--	The experiment object.

		Keyword arguments:
		string		--	An definition string. (default=None).
		"""

		if self.var is None:
			self.var = var_store(self, parent=experiment.var)
		self.name = name
		self.experiment = experiment
		self.debug = oslogger.debug_mode
		self.count = 0
		self._get_lock = None
		# Deduce item_type from class name
		prefix = self.experiment.item_prefix()
		self.item_type = str(self.__class__.__name__)
		if self.item_type.startswith(prefix):
			self.item_type = self.item_type[len(prefix):]
		if not hasattr(self, u'description'):
			self.var.description = self.default_description
		else:
			self.var.description = self.description
		self.from_string(string)

	def __deepcopy__(self, memo):
		"""Disable deep-copying of items"""
		
		return None

	@property
	def clock(self):
		return self.experiment._clock

	@property
	def log(self):
		return self.experiment._log

	@property
	def syntax(self):
		return self.experiment._syntax

	@property
	def python_workspace(self):
		return self.experiment._python_workspace

	@property
	def responses(self):
		return self.experiment._responses

	@property
	def default_description(self):
		return u'Default description'

	def reset(self):

		"""
		desc:
			Resets all item variables to their default value.
		"""

		pass

	def prepare(self):

		"""Implements the prepare phase of the item."""

		self.experiment.var.set(u'count_%s' % self.name, self.count)
		self.count += 1

	def run(self):

		"""Implements the run phase of the item."""

		pass

	def parse_variable(self, line):

		"""
		Reads a single variable from a single definition line.

		Arguments:
		line	--	A single definition line.

		Returns:
		True on succes, False on failure.
		"""

		# It is a little ugly to call parse_comment() here, but otherwise
		# all from_string() derivatives need to be modified
		if self.parse_comment(line):
			return True
		l = self.syntax.split(line.strip())
		if len(l) > 0 and l[0] == u'set':
			if len(l) != 3:
				raise osexception( \
					u'Error parsing variable definition: "%s"' % line)
			else:
				self.var.set(l[1], l[2])
				return True
		return False

	def parse_keywords(self, line, from_ascii=False, _eval=False):

		"""
		Parses keywords, e.g. 'my_keyword=my_value'.

		Arguments:
		line		--	A single definition line.

		Keyword arguments:
		from_ascii	--	DEPRECATED KEYWORD.
		_eval		--	Indicates whether the values should be evaluated.
						(default=False)

		Returns:
		A value dictionary with keywords as keys and values as values.
		"""

		# Parse keywords
		l = self.syntax.split(line.strip())
		keywords = {}
		for i in l:
			j = i.find(u'=')
			if j != -1:
				# UGLY HACK: if the string appears to be plain text,
				# rather than a keyword, for example something like
				# 'accuracy = [acc]%', do not parse it as a keyword-
				# value pair. The string needs to occur only once in
				# the full line, both quoted and unquoted.
				q = u'"%s"' % i
				if line.count(q) == 1 and line.count(i) == 1:
					oslogger.warning(
						u'"%s" does not appear to be a keyword-value pair in string "%s"'
						% (i, line)
					)
				else:
					var = str(i[:j])
					val = self.auto_type(i[j+1:])
					if _eval:
						val = self.syntax.eval_text(val)
					keywords[var] = val
		return keywords

	def parse_line(self, line):

		"""
		Allows for arbitrary line parsing, for item-specific requirements.

		Arguments:
		line	--	A single definition line.
		"""

		pass

	def parse_comment(self, line):

		"""
		Parses comments from a single definition line, indicated by # // or '.

		Arguments:
		line	--	A single definition line.

		Returns:
		True on succes, False on failure.
		"""

		line = line.strip()
		if len(line) > 0 and line[0] == u'#':
			self.comments.append(line[1:])
			return True
		elif len(line) > 1 and line[0:2] == u'//':
			self.comments.append(line[2:])
			return True
		return False

	def set_response(self, response=None, response_time=None, correct=None):

		"""
		desc:
			Deprecated by response store.
		"""

		warnings.warn(
			u'set_response() has been deprecated. Use responses object instead.',
			DeprecationWarning)
		self.responses.add(response=response, response_time=response_time,
			correct=correct)

	def __getattr__(self, var):

		if var in self.var.__vars__:
			warnings.warn(u'called %s as item property' % var,
				DeprecationWarning)
			return self.var.get(var)
		if hasattr(self.__class__, var):
			warnings.warn(
				u'called %s as item property and stored as class attribute' \
				% var, DeprecationWarning)
			return getattr(self.__class__, var)
		raise AttributeError(u'%s not found' % var)

	def variable_to_string(self, var):

		"""
		desc:
			Encodes a variable into a definition string.

		arguments:
			var:
				desc:	The name of the variable to encode.
				type:	str

		returns:
			desc:	A definition string.
			type:	str
		"""

		val = safe_decode(self.var.get(var, _eval=False))
		# Multiline variables are stored as a block
		if u'\n' in val or u'"' in val:
			s = u'__%s__\n' % var
			val = val.replace(u'__end__', u'\\__end__')
			for l in val.split(u'\n'):
				s += '\t%s\n' % l
			while s[-1] in (u'\t', u'\n'):
				s = s[:-1]
			s += u'\n\t__end__\n'
			return s
		# Regular variables
		return self.syntax.create_cmd(u'set', arglist=[var, val]) + u'\n'

	def from_string(self, string):

		"""
		desc:
			Parses the item from a definition string.

		arguments:
			string:
				desc:	A definition string, or None to reset the item.
				type:	[str, NoneType]
		"""

		self._script = string
		textblock_var = None
		self.var.clear()
		self.reset()
		self.comments = []
		if string is None:
			return
		for line in string.split(u'\n'):
			line_stripped = line.strip()
			# The end of a textblock
			if line_stripped == u'__end__':
				if textblock_var is None:
					raise osexception(u'It appears that a textblock has been '
						u'closed without being opened.')
				self.var.set(textblock_var,
					textblock_val.replace(u'\\__end__', u'__end__'))
				textblock_var = None
			# The beginning of a textblock. A new textblock is only started when
			# a textblock is not already ongoing, and only if the textblock
			# start is of the format __VARNAME__
			elif line_stripped[:2] == u'__' and line_stripped[-2:] == u'__' \
				and textblock_var is None:
				textblock_var = line_stripped[2:-2]
				if textblock_var != u'':
					textblock_val = u''
				else:
					textblock_var = None
				# We cannot just strip the multiline code, because that may mess
				# up indentation. So we have to detect if the string is indented
				# based on the opening __varname__ line.
				strip_tab = line[0] == u'\t'
			# Collect the contents of a textblock
			elif textblock_var is not None:
				if strip_tab:
					textblock_val += line[1:] + u'\n'
				else:
					textblock_val += line + u'\n'
			# Parse regular variables
			elif not self.parse_variable(line):
				self.parse_line(line)
		if textblock_var is not None:
			raise osexception(
				u'Missing __end__ block for multiline variable "%s" in item %s' \
				% (textblock_var, self.name))

	def to_string(self, item_type=None):

		"""
		Encodes the item into an OpenSesame definition string.

		Keyword arguments:
		item_type	--	The type of the item or None for autodetect.
						(default=None)

		Returns:
		The unicode definition string
		"""

		if item_type is None:
			item_type = self.item_type
		s = u'define %s %s\n' % (item_type, self.name)
		for comment in self.comments:
			s += u'\t# %s\n' % comment.strip()
		for var in self.var:
			s += u'\t' + self.variable_to_string(var)	
		return s

	def resolution(self):

		"""
		desc: |
			Returns the display resolution and checks whether the resolution is
			valid.

			__Important note:__

			The meaning of 'resolution' depends on the back-end. For example,
			the legacy back-end changes the actual resolution of the display,
			whereas the other back-ends do not alter the actual display
			resolution, but create a 'virtual display' with the requested
			resolution that is presented in the center of the display.

		returns:
			desc:	A (width, height) tuple
			type:	tuple
		"""

		w = self.var.get(u'width')
		h = self.var.get(u'height')
		if type(w) != int or type(h) != int:
			raise osexception( \
				u'(%s, %s) is not a valid resolution' % (w, h))
		return w, h

	def set(self, var, val):

		warnings.warn(u'item.set() is deprecated (for var %s)' % var,
			DeprecationWarning)
		setattr(self.var, var, val)

	def unset(self, var):

		warnings.warn(u'item.unset() is deprecated (for var %s)' % var,
			DeprecationWarning)
		self.var.unset(var)

	def get(self, var, _eval=True):

		warnings.warn(u'item.get() is deprecated (for var %s)' % var,
			DeprecationWarning)
		return self.var.get(var, _eval=_eval)

	def get_check(self, var, default=None, valid=None, _eval=True):

		warnings.warn(u'item.var.get() is deprecated (for var %s)' % var,
			DeprecationWarning)
		return self.var.get(var, default=default, _eval=_eval, valid=valid)

	def has(self, var):

		warnings.warn(u'item.has() is deprecated (for var %s)' % var,
			DeprecationWarning)
		return var in self.var

	def get_refs(self, text):

		"""
		desc:
			Returns a list of variables that are referred to by a string of
			text.

		arguments:
			text:
				desc:	A string of text. This can be any type, but will coerced
						to unicode if it is not unicode.

		returns:
			desc:	A list of variable names or an empty list if the string
					contains no references.
			type:	list

		Example: |
			print(self.get_refs('There are [two] [references] here'))
			# Prints ['two', 'references']
		"""

		text = safe_decode(text)

		l = []
		start = -1
		while True:
			# Find the start and end of a variable definition
			start = text.find(u'[', start + 1)
			if start < 0:
				break
			end = text.find(u']', start + 1)
			if end < 0:
				raise osexception( \
					u"Missing closing bracket ']' in string '%s', in item '%s'" \
					% (text, self.name))
			var = text[start+1:end]
			l.append(var)
			var = var[end:]
		return l

	def auto_type(self, val):

		"""
		desc:
			Converts a value into the 'best fitting' or 'simplest' type that is
			compatible with the value.

		arguments:
			val:
				desc:	A value. This can be any type.

		returns:
			desc:	The same value converted to the 'best fitting' type.
			type:	[unicode, int, float]

		Example: |
			print(type(self.auto_type('1'))) # Prints 'int'
			print(type(self.auto_type('1.1'))) # Prints 'float'
			print(type(self.auto_type('some text'))) # Prints 'unicode'
			# Note: Boolean values are converted to 'yes' / 'no' and are
			# therefore also returned as unicode objects.
			print(type(self.auto_type(True))) # Prints 'unicode'
		"""

		# Booleans are converted to True/ False
		if type(val) == bool:
			if val:
				return u'yes'
			else:
				return u'no'
		# Try to convert the value to a numeric type
		try:
			# Check if the value can be converted to an int without loosing
			# precision. If so, convert to int
			if int(float(val)) == float(val):
				return int(float(val))
			# Else convert to float
			else:
				return float(val)
		except:
			# Else, fall back to unicde
			return safe_decode(val)

	def set_item_onset(self, time=None):

		"""
		desc:
			Set a timestamp for the onset time of the item's execution.

		keywords:
			time:	A timestamp or None to use the current time.

		returns:
			desc:	A timestamp.
		"""

		if time is None:
			time = self.time()
		self.experiment.var.set(u'time_%s' % self.name, time)
		return time

	def dummy(self, **args):

		"""
		Dummy function

		Keyword arguments:
		arguments -- accepts all keywords for compatibility
		"""

		pass

	def var_info(self):

		"""
		Give a list of dictionaries with variable descriptions

		Returns:
		A list of (variable, description) tuples
		"""

		return [ (u"time_%s" % self.name, u"[Timestamp of last item call]"), \
			(u"count_%s" % self.name, u"[Number of item calls]") ]

	def sleep(self, ms):

		return self.clock.sleep(ms)

	def time(self):

		return self.clock.time()

	def flush_log(self):

		warnings.warn(u'item.flush_log() has been deprecated',
			DeprecationWarning)

	def split(self, s):

		warnings.warn(
			u'item.split() has been deprecated. Please use syntax.split()',
			DeprecationWarning)
		return self.syntax.split(s)

	def eval_text(self, text, round_float=False, soft_ignore=False,
		quote_str=False):

		warnings.warn(
			u'item.eval_text() has been deprecated. Please use syntax.eval_text()',
			DeprecationWarning)
		return self.syntax.eval_text(text, round_float=round_float)


Item = item



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets, QtGui
from libqtopensesame.widgets.base_widget import base_widget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'color_edit', category=u'core')


class color_edit(base_widget):

	"""
	desc:
		A colorpicker widget that emulates a QLineEdit.
	"""

	textChanged = QtCore.Signal('QString')
	textEdited = QtCore.Signal('QString')

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:
				desc:	The main-window object.
				type:	qtopensesame
		"""

		super(color_edit, self).__init__(main_window)
		self.edit = QtWidgets.QLineEdit()
		self._parent = None
		self.edit.setSizePolicy(QtWidgets.QSizePolicy.Minimum,
			QtWidgets.QSizePolicy.Minimum)
		self.edit.editingFinished.connect(self.apply)
		self.editingFinished = self.edit.editingFinished
		self.button = QtWidgets.QPushButton()
		self.button.setIconSize(QtCore.QSize(16,16))
		self.button.clicked.connect(self.colorpicker)
		layout = QtWidgets.QHBoxLayout()
		layout.setContentsMargins(0,0,0,0)
		layout.addWidget(self.edit)
		layout.addWidget(self.button)
		self.setLayout(layout)

	def colorpicker(self):

		"""
		desc:
			Picks a color with the colorpicker dialog.
		"""

		_color = QtWidgets.QColorDialog.getColor(QtGui.QColor(u'white'),
			self._parent, _(u'Pick a color'))
		if not _color.isValid():
			return
		self.setText(_color.name())
		self.apply()

	def text(self):

		"""
		desc:
			Gets text (emulate QLineEdit behavior).

		returns:
			desc:	A color text.
			type:	QString
		"""

		return self.edit.text()

	def setText(self, s):

		"""
		dsc:
			Sets text (emulate QLineEdit behavior).

		arguments:
			s:
				desc:	Text.
				type:	unicode
		"""

		self.edit.setText(safe_decode(s))

	def apply(self):

		"""
		desc:
			Emits a 'set_color' signal to indicate that a color has been picked.
		"""

		self.textChanged.emit(self.text())
		self.textEdited.emit(self.text())

	def initialize(self, experiment=None, color=None, parent=None):

		"""
		desc:
			Initializes the widget. This is necessary to apply the theme and
			give the fields initial values.

		keywords:
			experiment:
				desc:	The experiment object or None if it is already available
						via the base_component property.
				type:	[experiment, NoneType]
			color:
				color:	An initial color or None to start with experiment
						foreground.
				type:	[unicode, NoneType]
			parent:	A parent QWidget.
		"""

		if parent is not None:
			self._parent = parent
		else:
			self._parent = self.main_window
		if experiment is None:
			experiment = self.experiment
		if color is None:
			color = experiment.var.get(u'foreground', _eval=False)
		self.setText(color)
		self.button.setIcon(self.theme.qicon(u'os-color-picker'))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy.QtCore import QTimer
import multiprocessing
from distutils.version import StrictVersion
from libopensesame import metadata
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'update_checker', category=u'extension')


class update_checker(base_extension):

	def activate(self):

		self._check_for_updates()

	def event_startup(self):

		self._checking = False
		self._check_for_updates(always=False)

	def _check_for_updates(self, always=True):

		self._always = always
		if not always and not cfg.auto_update_check:
			oslogger.debug(u'skipping update check')
			return
		if self._checking:
			oslogger.debug(u'update check already in progress')
			return
		self._checking = True
		self._queue = multiprocessing.Queue()
		self._update_checker = multiprocessing.Process(
			target=_update_checker,
			args=(self._queue, cfg.remote_metadata_url)
		)
		self._update_checker.start()
		self.extension_manager.fire(
			'register_subprocess',
			pid=self._update_checker.pid,
			description='update_checker'
		)
		oslogger.debug(u'checking (PID={})'.format(self._update_checker.pid))
		QTimer.singleShot(1000, self._poll_update_process)

	def _error_notify(self):

		self.extension_manager.fire(
			u'notify',
			message=_(u'Failed to check for updates'),
			category=u'warning',
			always_show=True,
		)

	def _poll_update_process(self):

		if self._queue.empty():
			oslogger.debug(u'queue still empty')
			QTimer.singleShot(1000, self._poll_update_process)
			return
		content = self._queue.get()
		self._update_checker.join()
		try:
			self._update_checker.close()
		except AttributeError:
			# Process.close() was introduced only in Python 3.7
			pass
		self._checking = False
		if content is None:
			self._error_notify()
			return
		try:
			remote_metadata = safe_yaml_load(content)
			remote_strict_version = StrictVersion(remote_metadata[u'stable_version'])
		except Exception as e:
			self._error_notify()
			oslogger.error(u"failed to parse metadata: {}".format(e))
			return
		if remote_strict_version > metadata.strict_version:
			oslogger.debug(u"new version available")
			s = safe_read(self.ext_resource(u'update-available.md'))
			self.tabwidget.open_markdown(
				s % remote_metadata,
				title=u'Update available!'
			)
			return
		oslogger.debug(u"up to date")
		if not self._always:
			return
		self.tabwidget.open_markdown(
			self.ext_resource(u'up-to-date.md'),
			title=_(u'Up to date!')
		)


def _update_checker(queue, metadata_url):

	if py3:
		from urllib.request import urlopen
	else:
		from urllib2 import urlopen
	try:
		fd = urlopen(metadata_url)
		content = fd.read()
		fd.close()
	except Exception as e:
		print('failed to check for updates: {}'.format(e))
		queue.put(None)
	else:
		queue.put(content)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from qtpy import QtCore, QtWidgets
from libqtopensesame.dialogs.base_dialog import base_dialog

class text_input(base_dialog):

	"""
	desc:
		A simple text-input dialog.
	"""

	def __init__(self, main_window, msg=None, content=u'', validator=None):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	The main window object.
			msg:			A text message.

		keywords:
			content:		The starting content.
			validator:		A function to validate the input. This function
							should take a single str as argument, and return a
							bool.
		"""

		super(text_input, self).__init__(main_window,
			ui=u'dialogs.text_input_dialog')
		if msg is not None:
			self.ui.label_message.setText(msg)
		self._validator = validator
		self.ui.textedit_input.setPlainText(content)
		self.ui.textedit_input.setFont(self.experiment.monospace())
		self.adjustSize()

	def get_input(self):

		"""
		desc:
			Gets text input.

		returns:
			A string with text input or None of the dialog was not accepted.
		"""

		while True:
			if self.exec_() != QtWidgets.QDialog.Accepted:
				break
			s = self.ui.textedit_input.toPlainText()
			if self._validator is None or self._validator(s):
				return s
		return None



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._ellipse.ellipse import Ellipse


class Circle(Ellipse):

	def __init__(self, canvas, x, y, r, **properties):

		properties = properties.copy()
		properties.update({ 'x' : x, 'y' : y, 'w' : 2*r, 'h' : 2*r, u'r' : r })
		self.prepare = self.circle_prepare(self.prepare)
		Ellipse.__init__(self, canvas, **properties)

	@property
	def rect(self):
		return self.x-self.r, self.y-self.r, self.r*2, self.r*2

	def __contains__(self, xy):

		return ((xy[0]-self.x)**2+(xy[1]-self.y)**2)**0.5 <= self.r

	def circle_prepare(self, ellipse_prepare):

		"""
		desc:
			A decorator that converts the center coordinates used by the circle
			to the top-left coordinates used by the ellipse.
		"""

		def inner():

			r = self._properties[u'w'] // 2
			self._properties[u'x'] -= r
			self._properties[u'y'] -= r
			ellipse_prepare()
			self._properties[u'x'] += r
			self._properties[u'y'] += r

		return inner

	@staticmethod
	def _getter(key, self):

		if key == u'r':
			return self._properties[u'w'] // 2
		return Ellipse._getter(key, self)

	@staticmethod
	def _setter(key, self, val):

		if key == u'r':
			self._properties[u'w'] = val * 2
			self._properties[u'h'] = val * 2
			return
		Ellipse._setter(key, self, val)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
from libopensesame.widgets.themes.plain import plain
from openexp.canvas_elements import Rect, Line, Image
from openexp._canvas._element.group import Group


class gray(plain):

	def __init__(self, form):

		plain.__init__(self, form)
		self.box_checked_image = self.form.experiment.resource(os.path.join(
			u'widgets', u'gray', u'box-checked.png'))
		self.box_unchecked_image = self.form.experiment.resource(os.path.join(
			u'widgets', u'gray', u'box-unchecked.png'))

	def box(self, x, y, checked=False):

		return Image(
			self.box_checked_image if checked else self.box_unchecked_image,
			center=False, x=x, y=y).construct(self.canvas)

	def frame(self, x, y, w, h, style=u'normal'):

		elements = [
			Line(x+1, y+1, x+w-2, y+1, color=u'#babdb6'),
			Line(x+1, y+1, x+1, y+h-2, color=u'#babdb6'),
			Line(x+1, y+h-2, x+w-2, y+h-2, color=u'#555753'),
			Line(x+w-2, y+1, x+w-2, y+h-2, color=u'#555753')
			]
		if style in (u'active', u'normal'):
			elements.insert(0, Rect(x, y, w, h, color=u'#2e3436'))
		if style == u'normal':
			elements.append(
				Rect(x+2, y+2, w-4, h-4, color=u'#888a85', fill=True))
		return Group(self.canvas, elements)



#-*- coding:utf-8 -*-

"""
Created on 25-05-2012

Author: Edwin Dalmaijer

This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

"""

from libopensesame.py3compat import *
import pygame
from libopensesame import plugins


basejoystick = plugins.load_cls(
	__file__,
	cls=u'basejoystick',
	mod=u'basejoystick'
)


class legacy(basejoystick):

	def __init__(self, experiment, device=0, joybuttonlist=None, timeout=None):

		"""See _libjoystick.basejoystick"""

		self.js = pygame.joystick.Joystick(device)
		self.js.init()
		self.experiment = experiment
		self.set_joybuttonlist(joybuttonlist)
		self.set_timeout(timeout)
		pygame.event.set_blocked(pygame.JOYAXISMOTION)
		pygame.event.set_blocked(pygame.JOYHATMOTION)
		pygame.event.set_blocked(pygame.JOYBALLMOTION)

	def get_joybutton(self, joybuttonlist=None, timeout=None):

		"""See _libjoystick.basejoystick"""

		if joybuttonlist is None or joybuttonlist == []:
			joybuttonlist = self._joybuttonlist
		if timeout is None:
			timeout = self.timeout
		start_time = pygame.time.get_ticks()
		time = start_time
		while timeout is None or time - start_time <= timeout:
			time = pygame.time.get_ticks()
			for event in pygame.event.get():
				if event.type == pygame.KEYDOWN:
					if event.key == pygame.K_ESCAPE:
						self.experiment.pause()
				if event.type == pygame.JOYBUTTONDOWN:
					if (
						joybuttonlist is None or
						event.button + 1 in joybuttonlist
					):
						bpress = event.button + 1
						return bpress, time
		return None, time

	def get_joyaxes(self, timeout=None):

		"""See _libjoystick.basejoystick"""

		pygame.event.set_allowed(pygame.JOYAXISMOTION)
		if timeout is None:
			timeout = self.timeout
		pos = []
		start_time = pygame.time.get_ticks()
		time = start_time
		while timeout is None or time - start_time < timeout:
			time = pygame.time.get_ticks()
			for event in pygame.event.get():
				if event.type == pygame.KEYDOWN:
					if event.key == pygame.K_ESCAPE:
						self.experiment.pause()
				if event.type == pygame.JOYAXISMOTION:
					for axis in range(self.js.get_numaxes()):
						pos.append(self.js.get_axis(axis))
					pygame.event.set_blocked(pygame.JOYAXISMOTION)
					return pos, time
		pygame.event.set_blocked(pygame.JOYAXISMOTION)
		return None, time

	def get_joyballs(self, timeout=None):

		"""See _libjoystick.basejoystick"""

		pygame.event.set_allowed(pygame.JOYBALLMOTION)
		if timeout is None:
			timeout = self.timeout
		ballpos = []
		start_time = pygame.time.get_ticks()
		time = start_time
		while timeout is None or time - start_time < timeout:
			time = pygame.time.get_ticks()
			for event in pygame.event.get():
				if event.type == pygame.KEYDOWN:
					if event.key == pygame.K_ESCAPE:
						self.experiment.pause()
				if event.type == pygame.JOYBALLMOTION:
					for ball in range(self.js.get_numballs()):
						ballpos.append(self.js.get_ball(ball))
					pygame.event.set_blocked(pygame.JOYBALLMOTION)
					return ballpos, time
		pygame.event.set_blocked(pygame.JOYBALLMOTION)
		return None, time

	def get_joyhats(self, timeout=None):

		"""See _libjoystick.basejoystick"""

		pygame.event.set_allowed(pygame.JOYHATMOTION)
		if timeout is None:
			timeout = self.timeout
		hatpos = []
		start_time = pygame.time.get_ticks()
		time = start_time
		while timeout is None or time - start_time < timeout:
			time = pygame.time.get_ticks()
			for event in pygame.event.get():
				if event.type == pygame.KEYDOWN:
					if event.key == pygame.K_ESCAPE:
						self.experiment.pause()
				if event.type == pygame.JOYHATMOTION:
					for hat in range(self.js.get_numhats()):
						hatpos.append(self.js.get_hat(hat))
					pygame.event.set_blocked(pygame.JOYHATMOTION)
					return hatpos, time
		pygame.event.set_blocked(pygame.JOYHATMOTION)
		return None, time

	def get_joyinput(self, joybuttonlist=None, timeout=None):

		"""See _libjoystick.basejoystick"""

		pygame.event.set_allowed(pygame.JOYHATMOTION)
		pygame.event.set_allowed(pygame.JOYAXISMOTION)
		pygame.event.set_allowed(pygame.JOYBALLMOTION)
		if joybuttonlist is None or joybuttonlist == []:
			joybuttonlist = self._joybuttonlist
		if timeout is None:
			timeout = self.timeout
		pos = []
		ballpos = []
		hatpos = []
		eventtype = None
		start_time = pygame.time.get_ticks()
		time = start_time
		while timeout is None or time - start_time <= timeout:
			time = pygame.time.get_ticks()
			for event in pygame.event.get():
				if event.type == pygame.KEYDOWN:
					if event.key == pygame.K_ESCAPE:
						self.experiment.pause()
				if event.type == pygame.JOYBUTTONDOWN:
					if (
						joybuttonlist is None or
						event.button + 1 in joybuttonlist
					):
						eventtype = u'joybuttonpress'
						bpress = event.button + 1
						pygame.event.set_blocked(pygame.JOYHATMOTION)
						pygame.event.set_blocked(pygame.JOYAXISMOTION)
						pygame.event.set_blocked(pygame.JOYBALLMOTION)
						return eventtype, bpress, time
				if event.type == pygame.JOYAXISMOTION:
					eventtype = u'joyaxismotion'
					for axis in range(self.js.get_numaxes()):
						pos.append(self.js.get_axis(axis))
					pygame.event.set_blocked(pygame.JOYHATMOTION)
					pygame.event.set_blocked(pygame.JOYAXISMOTION)
					pygame.event.set_blocked(pygame.JOYBALLMOTION)
					return eventtype, pos, time
				if event.type == pygame.JOYBALLMOTION:
					eventtype = u'joyballmotion'
					for ball in range(self.js.get_numballs()):
						ballpos.append(self.js.get_ball(ball))
					pygame.event.set_blocked(pygame.JOYHATMOTION)
					pygame.event.set_blocked(pygame.JOYAXISMOTION)
					pygame.event.set_blocked(pygame.JOYBALLMOTION)
					return eventtype, ballpos, time
				if event.type == pygame.JOYHATMOTION:
					eventtype = u'joyhatmotion'
					for hat in range(self.js.get_numhats()):
						hatpos.append(self.js.get_hat(hat))
					pygame.event.set_blocked(pygame.JOYHATMOTION)
					pygame.event.set_blocked(pygame.JOYAXISMOTION)
					pygame.event.set_blocked(pygame.JOYBALLMOTION)
					return eventtype, hatpos, time
		pygame.event.set_blocked(pygame.JOYHATMOTION)
		pygame.event.set_blocked(pygame.JOYAXISMOTION)
		pygame.event.set_blocked(pygame.JOYBALLMOTION)
		return eventtype, None, time

	def input_options(self):

		"""See _libjoystick.basejoystick"""

		ninputs = [
			self.js.get_numbuttons(),
			self.js.get_numaxes(),
			self.js.get_numballs(),
			self.js.get_numhats()
		]
		return ninputs

	def flush(self):

		"""See _libjoystick.basejoystick"""

		joyinput = False
		for event in pygame.event.get():
			if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
				self.experiment.pause()
			if event.type in (
				pygame.JOYBUTTONDOWN,
				pygame.JOYAXISMOTION,
				pygame.JOYBALLMOTION,
				pygame.JOYHATMOTION
			):
				joyinput = True
		return joyinput



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.dialogs.base_dialog import base_dialog
from datamatrix import DataMatrix, operations
from qdatamatrix import QDataMatrix
import pickle
from qtpy import QtWidgets

class loop_wizard(base_dialog):

	"""
	desc:
		The loop-wizard dialog
	"""

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	The main window object.
		"""

		super(loop_wizard, self).__init__(main_window,
			ui=u'dialogs.loop_wizard_dialog')
		try:
			self._dm = pickle.loads(cfg.loop_wizard)
			assert(isinstance(self._dm, DataMatrix))
			assert(hasattr(self._dm._rowid, u'clone'))
		except:
			self._dm = DataMatrix(length=0)
		self._qdm = QDataMatrix(self._dm)
		self.ui.hbox_container.addWidget(self._qdm)
		self.ui.table_example.hide()

	def exec_(self):

		"""
		desc:
			Executes the dialog.

		returns:
			The dialog return status.
		"""

		retval = base_dialog.exec_(self)
		if retval == QtWidgets.QDialog.Rejected:
			return None
		self._dm = self._qdm.dm
		cfg.loop_wizard = pickle.dumps(self._dm)
		return operations.fullfactorial(self._dm)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._keyboard.legacy import *
import pygame
from openexp.backend import configurable
try:
	import android
except ImportError:
	android = None


class Droid(Legacy):

	"""
	desc: |
		This is a keyboard backend built on top of PyGame, adapted for Android
		devices.

		For function specifications and docstrings, see
		`openexp._keyboard.keyboard`.
	"""

	def _get_key_event(self, event_type):

		if not self.persistent_virtual_keyboard and android is not None:
			android.show_keyboard()
		start_time = pygame.time.get_ticks()
		time = start_time
		keylist = self.keylist
		timeout = self.timeout
		while True:
			time = pygame.time.get_ticks()
			for event in pygame.event.get(event_type):
				if event.key == pygame.K_ESCAPE:
					self.experiment.pause()
				# TODO The unicode mechanism that ensures compatibility between
				# keyboard layouts doesn't work for Android, so we use key
				# names. I'm not sure what effect this will have on non-QWERTY
				# virtual keyboards.
				if android is not None:
					key = pygame.key.name(event.key)
					if len(key) == 1 and (
						event.mod & pygame.KMOD_LSHIFT or
						event.mod & pygame.KMOD_RSHIFT
					):
						key = key.upper()
				else:
					# If we're not on Android, simply use the same logic as the
					# legacy back-end.
					if event.unicode in invalid_unicode:
						key = self.key_name(event.key)
					else:
						key = event.unicode
				if keylist is None or key in keylist:
					if not self.persistent_virtual_keyboard and android is not None:
						android.hide_keyboard()
					return key, time
			if timeout is not None and time-start_time >= timeout:
				break
			# Allow Android interrupt
			if android is not None and android.check_pause():
				android.wait_for_resume()
		if not self.persistent_virtual_keyboard and android is not None:
			android.hide_keyboard()
		return None, time

	def show_virtual_keyboard(self, visible=True):

		if android is None:
			return
		self.persistent_virtual_keyboard = visible
		if visible:
			android.show_keyboard()
		else:
			android.hide_keyboard()


# Non PEP-8 alias for backwards compatibility
droid = Droid
